""" Materials """

import os
import re
from typing import Any, Dict, List, Mapping, Tuple

import numpy as np
import pandas as pd
import tidy3d as td
from numpy.typing import NDArray
from pydantic.v1 import Field, validator
from scipy.constants import c
from scipy.ndimage import map_coordinates
from tidy3d import material_library

from .base_model import BaseModel, _array
from .environment import Environment

MATERIAL_TYPES: Dict[str, type] = {}


class Material(BaseModel):
    """a `Material` defines the refractive index of a `Structure3D` within an `Environment`."""

    type: str = ""

    name: str = Field(description="the name of the material")

    meta: Dict[str, Any] = Field(
        default_factory=lambda: {}, description="metadata for the material"
    )

    def __init_subclass__(cls):
        MATERIAL_TYPES[cls.__name__] = cls

    def __new__(cls, **kwargs):
        cls = MATERIAL_TYPES.get(kwargs.get("type", cls.__name__), cls)
        return BaseModel.__new__(cls)  # type: ignore

    @validator("type", pre=True, always=True)
    def validate_type(cls, value):
        if not value:
            value = getattr(cls, "__name__", "Geometry")
        if value not in MATERIAL_TYPES:
            raise ValueError(
                f"Invalid Material type. Got: {value!r}. Valid types: {MATERIAL_TYPES}."
            )
        return value

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        MATERIALS[self.name] = self

    def __call__(self, env: Environment) -> NDArray[np.complex_]:
        """returns an array of the refractive index at the wavelengths specified by the environment"""
        raise NotImplementedError("Please use one of the Material child classes")

    def _lumadd(self, sim, env, unit):
        from matplotlib.cm import get_cmap  # fmt: skip

        n = self(env)
        wl = np.asarray(env.wl * unit, dtype=complex).ravel()
        eps = np.asarray(n, dtype=complex).ravel() ** 2
        data = np.stack([c / wl, eps], 1)  # permittivity, not index

        if not sim.materialexists(self.name):
            sim.setmaterial(sim.addmaterial("Sampled data"), "name", self.name)
            color = np.asarray(
                self.meta.get("color") or get_cmap("jet")(np.abs(eps) / 15.0)
            )
            sim.setmaterial(self.name, "color", color)

        sim.setmaterial(self.name, "sampled data", data)

        return self.name

    class Config:
        fields = {
            "meta": {"exclude": True},
        }


class TidyMaterial(Material):
    name: str = Field(description="The material name as also used by tidy3d")
    variant: str = Field(description="The material name as also used by tidy3d")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # try:
        # # somehow black complains about the except
        # except ModuleNotFoundError:
        #    raise ModuleNotFoundError("Tidy3D has to be installed to use its material database")

        if self.name not in material_library:
            raise ValueError(
                f"Specified material name is invalid. Use one of {material_library.keys()}"
            )
        _material = material_library[self.name]
        _variants = getattr(_material, "variants", {})
        if not _variants:
            raise ValueError(f"Tidy3D material '{self.name}' not supported.")
        if self.variant not in _variants:
            _variant_options = list(_variants.keys())
            raise ValueError(
                f"Specified variant is invalid. Use one of {_variant_options}."
            )

    def __call__(self, env: Environment) -> NDArray[np.complex_]:
        if not isinstance(env, Environment):
            env = Environment(**env)
        _material = material_library[self.name]
        _variants = getattr(_material, "variants", None)
        assert _variants is not None
        mat = _material[self.variant]  # type: ignore
        eps_comp = getattr(mat, "eps_comp", None)
        assert eps_comp is not None
        eps = eps_comp(0, 0, td.C_0 / env.wl)
        return np.sqrt(eps)


class SampledMaterial(Material):
    params: Dict[str, np.ndarray[Tuple[int], np.dtype[np.float_]]] = Field(
        description="the wavelength over which the refractive index is defined."
    )
    n: np.ndarray[Tuple[int], np.dtype[np.complex_]] = Field(
        description="the complex refractive index of the material"
    )

    @staticmethod
    def _validate_array(arr):
        if isinstance(arr, dict):
            r = np.asarray(arr.get("real", 0.0), dtype=np.complex_)
            i = np.asarray(arr.get("imag", 0.0), dtype=np.complex_)
            new = r + 1j * i
        else:
            new = np.asarray(arr)
        return new.view(_array)

    @staticmethod
    def _validate_1d(name, arr):
        if arr.ndim != 1:
            raise ValueError(f"{name} should be 1D. Got a {arr.ndim}D array.")
        return arr

    @validator("n", pre=True)
    def validate_n_pre(cls, n):
        return cls._validate_1d("n", cls._validate_array(n))

    @validator("params", pre=True)
    def validate_params_pre(cls, params):
        if not isinstance(params, Mapping):
            raise TypeError(f"instance of dict expected. Got: {type(params)}.")
        return {
            k: cls._validate_1d(f"params:{k}", cls._validate_array(v))
            for k, v in params.items()
        }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        for p, v in self.params.items():
            Lp = v.shape[0]
            Ln = self.n.shape[0]
            if Lp != Ln:
                raise ValueError(
                    f"length of parameter array {p} does not match length of refractive index array n. \n"
                    f"{Lp} != {Ln}"
                )

    @classmethod
    def from_path(cls, path, meta=None):
        path = _validate_path(path)
        name = re.sub(r"\.csv$", "", os.path.split(path)[-1])
        df = pd.read_csv(path)
        return cls.from_df(name, df, meta=meta)

    @classmethod
    def from_df(cls, name, df, meta=None):
        meta = meta or {}

        nr = df["nr"].values
        ni = np.zeros_like(nr) if "ni" not in df else df["ni"].values
        n = nr + 1j * ni

        columns = [c for c in df.columns if c not in ["nr", "ni"]]
        params = {c: np.asarray(df[c].values, dtype=np.float_) for c in columns}

        return cls(name=name, params=params, n=n, meta=meta)

    def __call__(self, env: Environment) -> NDArray[np.complex_]:
        if not isinstance(env, Environment):
            env = Environment(**env)
        # n = interp1d(1 / self.params['wl'], self.n, fill_value="extrapolate")(1 / env.wl)
        df = pd.DataFrame({**self.params, "nr": np.real(self.n), "ni": np.imag(self.n)})
        data, params, strings = _to_ndgrid(df, wl_key="wl")
        result, axs, pos = _evaluate_general_corner_model(
            data,
            params,
            strings,
            **{k: np.atleast_1d(v) for k, v in env.dict().items()},
        )
        nr = result.take(pos["targets"]["nr"], axs["targets"])
        ni = result.take(pos["targets"]["ni"], axs["targets"])
        n = nr + 1j * ni
        # FIXME: ideally we should just be able to return `n` here...
        # return n
        return np.squeeze(np.real(n))  # TODO: allow complex multi-dimensional n


Materials = List[Material]
MATERIALS: Dict[str, Material] = {}


def _to_ndgrid(df, wl_key="wl"):
    """convert stacked data to hypercube data

    Args:
        df: the dataframe to convert to hypercube data
        wl_key: which key to use to sort. Options: 'wl' or 'f'. Using 'f' yields better interpolation results
            as most optical phenomina are more linear in 'f' than in 'wl'.

    Note:
        This only works if the number of other parameters (e.g. # wls) stays consistent for each corner.
    """
    df = df.copy()
    if wl_key not in ["f", "wl"]:
        raise ValueError(
            f"Unsupported wl_key. Valid choices: 'wl', 'f'. Got: {wl_key}."
        )
    df["f"] = c / df["wl"].values
    value_columns = [c for c in ["nr", "ni"] if c in df.columns]
    param_columns = [c for c in df.columns if c not in value_columns + ["wl", "f"]] + [
        wl_key
    ]
    param_columns = [c for c in param_columns if c in df.columns]
    df = _sort_rows(
        df[param_columns + value_columns], not_by=value_columns, wl_key=wl_key
    )
    params = {c: df[c].unique() for c in param_columns}
    data = df[value_columns].values
    data = data.reshape(*(v.shape[0] for v in params.values()), len(value_columns))
    data = np.asarray(data)
    params["targets"] = np.array(value_columns, dtype=object)
    params, strings = _extract_strings(params)
    return data, params, strings


def _evaluate_general_corner_model(data, params, strings, /, **kwargs):
    given_params = {k: kwargs.get(k, np.asarray(v).mean()) for k, v in params.items()}
    given_strings = {
        p: (list(v.values()) if p not in kwargs else [v[vv] for vv in kwargs[p]])
        for p, v in strings.items()
    }
    string_locations = {k: i for i, k in enumerate(given_params) if k in given_strings}
    param_locations = {
        k: i for i, k in enumerate(given_params) if k not in given_strings
    }
    data = np.transpose(
        data, axes=[*param_locations.values(), *string_locations.values()]
    )
    _, string_shape = (
        data.shape[: len(param_locations)],
        data.shape[len(param_locations) :],
    )
    given_params = {k: v for k, v in given_params.items() if k not in given_strings}
    num_params = len(given_params)
    data = _downselect(data, given_strings.values())
    param_shape, string_shape = data.shape[:num_params], data.shape[num_params:]
    data = data.reshape(*param_shape, -1)
    given_param_values = [np.asarray(v) for v in given_params.values()]
    stacked_params = np.stack(np.broadcast_arrays(*given_param_values), 0)
    coords = _get_coordinates([v for k, v in params.items() if k in given_params], stacked_params)  # type: ignore
    result = _map_coordinates(data, coords)
    axs = {k: i for i, k in enumerate(given_strings)}
    rev_strings = {k: {vv: kk for kk, vv in v.items()} for k, v in strings.items()}
    pos = {
        k: {rev_strings[k][vv]: i for i, vv in enumerate(v)}
        for k, v in given_strings.items()
    }
    return result.reshape(*string_shape, *result.shape[1:]), axs, pos


def _downselect(data, idxs_list):
    for i, idxs in enumerate(reversed(idxs_list), start=1):
        data = data.take(np.asarray(idxs, dtype=int), axis=-i)
    return data


# @partial(jax.vmap, in_axes=(-1, None), out_axes=0)
# def _map_coordinates(input, coordinates):
#    return jax.scipy.ndimage.map_coordinates(input, coordinates, 1, mode="nearest")


def _map_coordinates(data: np.ndarray, coordinates):
    result = []
    for i in range(data.shape[-1]):
        current_result = map_coordinates(
            data[..., i], coordinates, order=1, mode="nearest"
        )
        result.append(current_result)
    return np.stack(result, 0)


def _get_coordinate(arr1d: np.ndarray, value: np.ndarray):
    return np.interp(value, arr1d, np.arange(arr1d.shape[0]))


def _get_coordinates(arrs1d: List[np.ndarray], values: np.ndarray):
    # don't use vmap as arrays in arrs1d could have different shapes...
    return np.array([_get_coordinate(a, v) for a, v in zip(arrs1d, values)])


def _extract_strings(params):
    string_map = {}
    new_params = {}
    for k, v in list(params.items()):
        if v.dtype == object:  # probably string array!
            string_map[k] = {s: i for i, s in enumerate(v)}
            new_params[k] = np.array(list(string_map[k].values()), dtype=int)
        else:
            new_params[k] = np.array(v, dtype=float)
    return new_params, string_map


def _sort_rows(df, not_by=("nr", "ni"), wl_key="wl"):
    not_by = ["wl", "f"] + list(not_by)
    by = [c for c in df.columns if c not in not_by] + [wl_key]
    return df.sort_values(by=by).reset_index(drop=True)


def _validate_path(path):
    if not path.endswith(".csv"):
        path = f"{path}.csv"
    path_parts = path.replace("\\", "/").split("/")
    if not os.path.exists(path) and len(path_parts) == 1:
        lib_path = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(lib_path, "assets", path)
    if not os.path.exists(path):
        raise FileNotFoundError(f"Material file {path!r} not found.")
    path = os.path.relpath(path, start=os.getcwd())
    return path


silicon = SampledMaterial.from_path(
    path="silicon",
    meta={"color": (0.9, 0, 0, 0.9)},
)
silicon_oxide = SampledMaterial.from_path(
    path="silicon_oxide",
    meta={"color": (0.9, 0.9, 0.9, 0.9)},
)

silicon_nitride = TidyMaterial(name="Si3N4", variant="Luke2015PMLStable")



""" GDS Extrusions """

from typing import Dict, List, Tuple

import numpy as np
import shapely.geometry as sg
from pydantic.v1 import Field

from .base_model import BaseModel
from .geometries import Prism
from .materials import Material
from .structures import Structure3D

# TODO: Maybe it makes more sense to use native GDSFactory tooling for this


class GdsExtrusionRule(BaseModel):
    """a `GdsExtrusionRule` describes a single extrusion rule.
    Multiple of such rules can later be associated with a gds layer tuple."""

    material: Material = Field(description="the material of the extrusion")
    h_min: float = Field(description="the extrusion starting height")
    h_max: float = Field(description="the extrusion ending height")
    buffer: float = Field(
        default=0.0,
        description="an extra buffer (=grow or shrink) operation applied to the polygon",
    )
    mesh_order: int = Field(
        default=5.0, description="the mesh order of the resulting `Structure3D`"
    )

    def __call__(self, poly) -> Structure3D:
        if self.buffer > 0:
            try:
                poly = np.asarray(sg.Polygon(poly).buffer(self.buffer).boundary.coords)
            except NotImplementedError:
                import gdspy  # fmt: skip

                polygonset = gdspy.offset(gdspy.Polygon(poly), 0.25)
                assert polygonset is not None
                poly = polygonset.polygons[0]
        return Structure3D(
            material=self.material,
            geometry=Prism(
                poly=poly,
                h_min=self.h_min,
                h_max=self.h_max,
                axis="y",
            ),
            mesh_order=self.mesh_order,
        )


def extrude_gds(
    cell,  # type: ignore
    extrusions: Dict[Tuple[int, int], List[GdsExtrusionRule]],
):
    """extrude a gds cell given a dictionary of extruson rules

    Args:
        cell: a gdspy or gdstk Cell to extrude
        extrusions: the extrusion rules to use (if not given, the example extrusions will be used.)
    """
    structs = []
    for layer, polys in cell.get_polygons(by_spec=True, depth=None).items():
        for poly in polys:
            if layer not in extrusions:
                continue
            structs.extend(extrusion(poly) for extrusion in extrusions[layer])
    return structs



""" A simulation Environment """

from pydantic.v1 import Extra, Field

from .base_model import BaseModel


class Environment(BaseModel):
    """An environment contains all variables that don't depend on the
    geometry/structure itself such as most commonly wavelength and temperature."""

    wl: float = Field(default=1.5, description="the wavelength of the environment")
    T: float = Field(default=25.0, description="the temperature of the environment")

    class Config:
        allow_population_by_field_name = True
        extra = Extra.allow



""" Provides a custom pydantic BaseModel which handles numpy arrays better """

from functools import wraps
from hashlib import md5
from typing import Any, Tuple

import numpy as np
from pydantic.v1.main import BaseModel as _BaseModel
from pydantic.v1.main import ModelMetaclass, PrivateAttr

from .cache import cache_array, cache_model


class _array(np.ndarray):
    """just an immutable array with a nicer repr"""

    def __new__(cls, *args, **kwargs):
        arr = np.ndarray.__new__(cls, *args, **kwargs)
        arr.setflags(write=False)
        return arr

    def __hash__(self):
        return (
            int.from_bytes(md5(self.tobytes()).digest(), byteorder="big")
            % 1000000000000
        )

    def __repr__(self):
        if self.ndim == 0:
            return f"{self:.3e}"
        cls_str = f"array{{{'x'.join(str(i) for i in self.shape)}}}"
        self = self.ravel()
        num_els = self.shape[0]
        if num_els == 0:
            return f"{cls_str}([])"
        elif num_els == 1:
            return f"{cls_str}([{self[0]:.3e}])"
        elif num_els == 2:
            return f"{cls_str}([{self[0]:.3e}, {self[1]:.3e}])"
        elif num_els == 3:
            return f"{cls_str}([{self[0]:.3e}, {self[1]:.3e}, {self[2]:.3e}])"
        else:
            return f"{cls_str}([{self[0]:.3e}, {self[1]:.3e}, ..., {self[-1]:.3e}])"


def _array_cls(annot) -> type:
    def get_validators(cls) -> Any:
        yield cls.validate

    def validate(cls, x) -> _array:
        if isinstance(x, dict):
            r = np.asarray(x.get("real", 0.0), dtype=np.float_)
            i = np.asarray(x.get("imag", 0.0), dtype=np.float_)
            x = r + 1j * i

        shape, dtype = _parse_array_type_info(cls.annot)
        x = np.asarray(x, dtype=(None if dtype is Any else dtype))
        if shape is not Any:
            try:
                shape = np.broadcast_shapes(shape, x.shape)
            except ValueError:
                raise ValueError(
                    f"Invalid shape for attribute 'x': Expected: {shape}. Got: {x.shape}."
                )
            x = np.broadcast_to(x, shape)

        return x.view(_array)

    def modify_schema(cls, schema, field):
        schema["title"] = field.name.replace("_", " ").title()
        schema["default"] = "array"

    Array = type(
        "Array",
        (np.ndarray,),
        {
            "annot": annot,
            "__get_validators__": classmethod(get_validators),
            "validate": classmethod(validate),
            "__modify_schema__": classmethod(modify_schema),
        },
    )
    return Array


def _serialize_array(arr):
    if np.iscomplexobj(arr):
        return {
            "real": _serialize_array(np.real(arr)),
            "imag": _serialize_array(np.imag(arr)),
        }
    else:
        return np.round(arr, 16).tolist()


def _parse_array_type_info(annotation) -> Tuple[Any, Any]:
    try:
        type_info = annotation.__args__
    except AttributeError:
        type_info = tuple()
    if len(type_info) == 1:
        (shape,) = type_info
        dtype = Any
    elif len(type_info) == 2:
        shape, dtype = type_info
    else:
        shape, dtype = Any, Any
    try:
        (dtype,) = dtype.__args__  # type: ignore
    except Exception:
        dtype = Any

    try:
        shape = shape.__args__  # type: ignore
        shape = tuple(_try_parse_shape_int(i) for i in shape)
    except Exception:
        shape = Any

    return shape, dtype


def _try_parse_shape_int(value):
    from typing import _LiteralGenericAlias  # fmt: skip # type: ignore

    if isinstance(value, _LiteralGenericAlias):
        (value,) = value.__args__
    try:
        return int(value)
    except Exception:
        return 1


class _ModelMetaclass(ModelMetaclass):
    """A metaclass (used in our custom BaseModel) to handle numpy array type hints better.

    Generic numpy type hints (will) have the following syntax::

        arr: np.ndarray[Shape, DType]

    for example::

        arr: np.ndarray[Tuple[Literal[1], int], np.dtype[np.float64]] = np.array([[1, 2, 3]])

    When using this _ModelMetaclass, the BaseModel will automatically try to cast the arrays according to the given array type-hint.

    Moreover, the following config will be injected (added to the config class that might already be present)::

        Config:
            arbitrary_types_allowed = True
            json_encoders = {
                np.ndarray: lambda arr: np.round(arr, 16).tolist()
            }

    """

    def __new__(cls, name, bases, dct, **kwargs):
        # Inject sensible default Configuration (Config class)
        extra_config = {
            "allow_mutation": False,
            "arbitrary_types_allowed": True,
            "json_encoders": {
                np.ndarray: _serialize_array,
                _array: _serialize_array,
                complex: lambda value: {"real": np.real(value), "imag": np.imag(value)},
            },
        }
        if "Config" not in dct:
            dct["Config"] = type("Config", (), extra_config)
        else:
            config = {**dct["Config"].__dict__}
            config["arbitrary_types_allowed"] = True
            config["json_encoders"] = {
                **config.get("json_encoders", {}),
                **extra_config["json_encoders"],
            }
            dct["Config"] = type("Config", (dct["Config"],), config)

        # Enforce numpy array annotations
        annotations = dct.get("__annotations__", {})
        for attr, annot in annotations.items():
            if cls._is_array_annot(annot):
                annotations[attr] = _array_cls(annot)
            if annot is complex:
                annotations[attr] = _array_cls(np.ndarray[Any, np.dtype[np.complex_]])
        return super().__new__(cls, name, bases, dct, **kwargs)

    @staticmethod
    def _is_array_annot(annot):
        try:
            arr = annot(0)
        except Exception:
            return False
        return isinstance(arr, np.ndarray)


class BaseModel(_BaseModel, metaclass=_ModelMetaclass):
    """A customized pydantic base model that handles numpy array type hints"""

    _cache = PrivateAttr({})

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__dict__.update({k: _view_arrays(k, v) for k, v in self.__dict__.items()})
        model = cache_model(self)
        if not model is self:
            object.__setattr__(self, "__dict__", model.__dict__)

    def _repr(self, indent=0, shift=2):
        s = f"{self.__class__.__name__}("
        dic = self.dict()
        if dic:
            s = s + "\n"
        for key in dic:
            if key == "data":
                continue
            attr = getattr(self, key)
            if isinstance(attr, BaseModel):
                attr_str = attr._repr(indent=indent + shift, shift=shift)
            else:
                attr_str = repr(attr)
            s += f"{' '*(indent + shift)}{key}={attr_str}\n"
        s += f"{' '*indent})"
        return s

    def _visualize(self):
        raise NotImplementedError(
            f"visualization for {self.__class__.__name__!r} not (yet) implemented."
        )

    def __hash__(self):
        try:
            arr = np.frombuffer(md5(self.json().encode()).digest(), dtype=np.uint8)[-8:]
            idx = np.arange(arr.shape[0], dtype=np.int64)[::-1]
            return int(np.sum(arr * 255**idx))
        except Exception:
            return None

    def __repr__(self):
        return self._repr()

    def __str__(self):
        return self._repr()


def _view_arrays(key, obj):
    if isinstance(obj, dict):
        return {k: _view_arrays(k, v) for k, v in obj.items()}
    elif isinstance(obj, np.ndarray) or isinstance(obj, _array):
        obj = obj.view(_array)

        # arbitrary: let's not spam the cache with this.
        if not key in ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]:
            obj = cache_array(obj)

        return obj
    else:
        return obj


def cache(prop):
    prop_name = prop.__name__

    @wraps(prop)
    def getter(self):
        stored_value = self._cache.get(prop_name)

        if stored_value is not None:
            return stored_value

        computed = prop(self)
        self._cache[prop_name] = computed
        return computed

    return getter


def cached_property(method):
    return property(cache(method))



""" EME Implementations & Backends """

from functools import wraps

from .common import compute_interface_s_matrices as compute_interface_s_matrices
from .common import compute_interface_s_matrix as compute_interface_s_matrix
from .common import compute_propagation_s_matrices as compute_propagation_s_matrices
from .common import compute_propagation_s_matrix as compute_propagation_s_matrix
from .common import select_ports as select_ports
from .sax import compute_s_matrix_sax as compute_s_matrix_sax


@wraps(compute_s_matrix_sax)
def compute_s_matrix(*args, **kwargs):
    return compute_s_matrix_sax(*args, **kwargs)



""" FDE Lumerical Backend """

import numpy as np
from pydantic.v1.types import PositiveInt

from ..cross_section import CrossSection
from ..environment import Environment
from ..mode import Mode, normalize_product, zero_phase
from ..structures import Structure3D

_global = {"sim": None}


def get_sim(**kwargs):
    sim = kwargs.get("sim", None)
    if sim is not None:
        return sim
    sim = _global["sim"]
    if sim is None:
        raise ValueError(
            "Could not start Lumerical simulation. "
            "Please either pass the `lumapi.MODE` simulation object as an argument to `compute_modes_lumerical` or "
            "use `set_sim(sim)` to globally set the lumapi.MODE simulation object."
        )
    return sim


def create_lumerical_geometries(
    sim, structures: list[Structure3D], env: Environment, unit: float
):
    sim = get_sim(sim=sim)
    sim.switchtolayout()
    sim.deleteall()
    for s in structures:
        s._lumadd(sim, env, unit, "yzx")


# @validate_arguments
def compute_modes_lumerical(
    cs: CrossSection,
    num_modes: PositiveInt = 10,
    sim=None,
    unit: float = 1e-6,
):
    """compute ``Modes` for a given ``FdeSpec` (Lumerical backend)

    Args:
        sim: the lumerical simulation object
        spec: The FDE simulation specification
        unit: Conversion factor between MEOW unit (probably um) and Lumerical unit (probably m).
    """
    from lumapi import LumApiError  # fmt: skip # type: ignore

    sim = get_sim(sim=sim)
    cell = cs._cell
    assert cell is not None

    _assert_default_mesh_setting(cell.mesh.angle_phi == 0, "angle_phi")
    _assert_default_mesh_setting(cell.mesh.angle_theta == 0, "angle_theta")
    _assert_default_mesh_setting(cell.mesh.bend_radius is None, "bend_radius")

    assert sim is not None
    create_lumerical_geometries(sim, cell.structures, cs.env, unit)
    sim.select("FDE")
    sim.delete()
    pml_settings = {}
    num_pml_y, num_pml_z = 0, 0
    if cell.mesh.num_pml[0] > 0:
        pml_settings.update(
            {
                "y_min_bc": "PML",
                "y_max_bc": "PML",
            }
        )
        num_pml_y = 22  # TODO: allow adjusting these values
    if cell.mesh.num_pml[1] > 0:
        pml_settings.update(
            {
                "z_min_bc": "PML",
                "z_max_bc": "PML",
            }
        )
        num_pml_z = 22  # TODO: allow adjusting these values
    sim.addfde(
        background_index=1.0,
        solver_type="2D X normal",
        x=float(cell.z * unit),
        y_min=float(cell.mesh.x.min() * unit),
        y_max=float(cell.mesh.x.max() * unit),
        z_min=float(cell.mesh.y.min() * unit),
        z_max=float(cell.mesh.y.max() * unit),
        define_y_mesh_by="number of mesh cells",
        define_z_mesh_by="number of mesh cells",
        mesh_cells_y=cell.mesh.x_.shape[0],
        mesh_cells_z=cell.mesh.y_.shape[0],
        **pml_settings,
    )
    # set mesh size again, because PML messes with it:
    if cell.mesh.num_pml[0] > 0:
        sim.setnamed("FDE", "mesh cells y", cell.mesh.x_.shape[0] - num_pml_y)
    if cell.mesh.num_pml[1] > 0:
        sim.setnamed("FDE", "mesh cells z", cell.mesh.y_.shape[0] - num_pml_z)
    sim.setanalysis("number of trial modes", int(num_modes))
    sim.setanalysis("search", "near n")
    sim.setanalysis("use max index", True)
    sim.setanalysis("wavelength", float(cs.env.wl * unit))
    sim.findmodes()
    modes = []
    for j in range(1, num_modes + 1):
        try:
            mode = _lumerical_fields_to_mode(
                cs=cs,
                lneff=sim.getdata(f"mode{j}", "neff"),
                lEx=sim.getdata(f"mode{j}", "Ex"),
                lEy=sim.getdata(f"mode{j}", "Ey"),
                lEz=sim.getdata(f"mode{j}", "Ez"),
                lHx=sim.getdata(f"mode{j}", "Hx"),
                lHy=sim.getdata(f"mode{j}", "Hy"),
                lHz=sim.getdata(f"mode{j}", "Hz"),
            )
        except LumApiError:
            break
        mode = normalize_product(mode)
        mode = zero_phase(mode)
        modes.append(mode)

    return sorted(modes, key=lambda m: np.real(m.neff), reverse=True)


def _lumerical_fields_to_mode(cs, lneff, lEx, lEy, lEz, lHx, lHy, lHz):
    return Mode(
        cs=cs,
        neff=lneff.ravel().item(),
        Ex=lEy.squeeze()[1:, :-1],
        Ey=lEz.squeeze()[:-1, 1:],
        Ez=lEx.squeeze()[:-1, :-1],
        Hx=lHy.squeeze()[:-1, 1:],
        Hy=lHz.squeeze()[1:, :-1],
        Hz=lHx.squeeze()[1:, 1:],
    )


def _assert_default_mesh_setting(condition, param_name):
    if not condition:
        raise NotImplementedError(
            f"Setting mesh.{param_name} is currently not supported in the Lumerical Backend. "
            "Please open an issue of submit a PR on GitHub to fix this: ",
            "https://github.com/flaport/meow",
        )


def set_sim(sim):
    _global["sim"] = sim



""" An EigenMode """

import pickle
import warnings
from itertools import product
from typing import List, Literal, Optional, Tuple

import numpy as np
from pydantic.v1 import Field
from scipy.constants import epsilon_0 as eps0
from scipy.constants import mu_0 as mu0
from scipy.linalg import norm

from .base_model import BaseModel, cached_property
from .cross_section import CrossSection
from .integrate import integrate_2d
from .visualize import _figsize_visualize_mode


class Mode(BaseModel):
    """A `Mode` contains the field information for a given `CrossSection`."""

    neff: complex = Field(description="the effective index of the mode")
    cs: CrossSection = Field(
        description="the index cross section for which the mode was calculated"
    )
    Ex: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Ex-fields of the mode"
    )
    Ey: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Ey-fields of the mode"
    )
    Ez: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Ez-fields of the mode"
    )
    Hx: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Hx-fields of the mode"
    )
    Hy: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Hy-fields of the mode"
    )
    Hz: np.ndarray[Tuple[int, int], np.dtype[np.complex_]] = Field(
        description="the Hz-fields of the mode"
    )
    interpolation: Optional[Literal["Ex", "Ey", "Ez", "Hz"]] = Field(
        default=None,
        description="To which 2D Yee-location the fields are interpolated to.",
    )

    def interpolate(self, location: Literal["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]):
        if self.interpolation is not None:
            raise RuntimeError("Cannot interpolate from already interpolated mode!")
        interpolate_funcs = {
            "EX": _interpolate_Ex,
            "EY": _interpolate_Ey,
            "EZ": _interpolate_Ez,
            "HX": _interpolate_Ey,
            "HY": _interpolate_Ex,
            "HZ": _interpolate_Hz,
        }
        interpolate_func = interpolate_funcs[location.upper()]
        return interpolate_func(self)

    @property
    def te_fraction(self):
        """the TE polarization fraction of the mode."""
        return te_fraction(self)

    @cached_property
    def _pointing(self):
        """calculate and cache the poynting vector"""
        vecE = np.stack([self.Ex, self.Ey, self.Ez], axis=-1)
        vecH = np.stack([self.Hx, self.Hy, self.Hz], axis=-1)
        vecP = np.cross(vecE, vecH)
        Px, Py, Pz = np.rollaxis(vecP, -1)
        return {
            "Px": Px,
            "Py": Py,
            "Pz": Pz,
        }

    @property
    def Px(self):
        return self._pointing["Px"]

    @property
    def Py(self):
        return self._pointing["Py"]

    @property
    def Pz(self):
        return self._pointing["Pz"]

    @cached_property
    def A(self):
        """mode area"""
        vecE = np.stack([self.Ex, self.Ey, self.Ez], axis=-1)
        E_sq = norm(vecE, axis=-1, ord=2)
        E_qu = E_sq**2
        x = self.cs.mesh.x_
        y = self.cs.mesh.y_
        return np.float_(integrate_2d(x, y, E_sq) ** 2 / integrate_2d(x, y, E_qu))

    @property
    def env(self):
        return self.cs.env

    @property
    def mesh(self):
        return self.cs.mesh

    def _visualize(
        self,
        title=None,
        title_prefix="",
        fields=None,
        ax=None,
        n_cmap=None,
        mode_cmap=None,
        num_levels=8,
        operation=lambda x: np.abs(x) ** 2,
        show=True,
    ):
        import matplotlib.pyplot as plt  # fmt: skip
        from matplotlib import colors  # fmt: skip
        from mpl_toolkits.axes_grid1 import make_axes_locatable  # fmt: skip
        W, H = _figsize_visualize_mode(self.cs, 6.4)

        if not fields:
            fields = ["Ex"]

        if len(fields) > 1:
            if len(fields) > 2:
                max_width = 15
                current_width = len(fields) * W
                W, H = _figsize_visualize_mode(self.cs, 6.4 * max_width / current_width)
            if ax is None:
                _, ax = plt.subplots(1, len(fields), figsize=(len(fields) * W, H))
            if len(ax) < len(fields):
                raise ValueError(
                    f"Not enough axes supplied for the number of fields "
                    f"to plot! {len(ax)} < {len(fields)}."
                )
            for field, ax_ in zip(fields, ax):
                self._visualize(
                    title=title,
                    title_prefix=title_prefix,
                    fields=[field],
                    ax=ax_,
                    n_cmap=n_cmap,
                    mode_cmap=mode_cmap,
                    num_levels=num_levels,
                    operation=operation,
                    show=False,
                )
            if show:
                plt.show()
            return

        field = fields[0]
        valid_fields = ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz", "Px", "Py", "Pz"]
        if field not in valid_fields:
            raise ValueError(
                f"Invalid field {field!r}. Valid fields: {', '.join(valid_fields)}."
            )

        if ax is None:
            ax = plt.gca()
        plt.sca(ax)

        if n_cmap is None:
            # little bit lighter colored than the one in cs._visualize:
            n_cmap = colors.LinearSegmentedColormap.from_list(
                name="c_cmap", colors=["#ffffff", "#c1d9ed"]
            )
        self.cs._visualize(ax=ax, n_cmap=n_cmap, cbar=False, show=False)

        x, y = "x", "y"  # currently only propagation in z supported, see Mesh2D
        c = {
            "Ex": "x",
            "Ey": "y",
            "Ez": "z",
            "Hx": "y",
            "Hy": "x",
            "Hz": "z_",
            "Px": "x",
            "Py": "y",
            "Pz": "z",
        }[field]
        if mode_cmap is None:
            mode_cmap = "inferno"
        X = getattr(self.mesh, f"X{c}")
        Y = getattr(self.mesh, f"Y{c}")
        mode = operation(getattr(self, field))
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            levels = np.linspace(mode.min(), mode.max(), num_levels + 1)[1:]
            plt.contour(X, Y, mode, cmap=mode_cmap, levels=levels)  # fmt: skip
            # plt.pcolormesh(X, Y, mode, cmap=mode_cmap, alpha=0.5) #, levels=levels)  # fmt: skip
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(cax=cax)
        plt.sca(ax)

        plt.xlabel(x)
        plt.ylabel(y)
        plt.grid(True, alpha=0.4)
        if title is None:
            plt.title(f"{title_prefix}{field} [neff={float(np.real(self.neff)):.6f}]")
        else:
            plt.title(f"{title_prefix}{title}")
        plt.xlim(X.min(), X.max())
        plt.ylim(Y.min(), Y.max())
        plt.axis("scaled")
        if show:
            plt.show()

    def save(self, filename):
        with open(filename, "wb") as file:
            pickle.dump(self, file)

    @classmethod
    def load(cls, filename):
        with open(filename, "rb") as file:
            return pickle.load(file)

    def __add__(self, other):
        if not isinstance(other, Mode):
            raise TypeError(
                f"unsupported operand type(s) for +: 'Mode' and '{type(other).__name__}'"
            )
        new_mode = Mode(
            neff=np.mean([self.neff, other.neff]),
            cs=self.cs,
            Ex=self.Ex + other.Ex,
            Ey=self.Ey + other.Ey,
            Ez=self.Ez + other.Ez,
            Hx=self.Hx + other.Hx,
            Hy=self.Hy + other.Hy,
            Hz=self.Hz + other.Hz,
        )
        return new_mode

    def __mul__(self, other):
        if not isinstance(other, (float, np.floating, complex, int, np.integer)):
            raise TypeError(
                f"unsupported operand type(s) for *: 'Mode' and '{type(other).__name__}'"
            )
        new_mode = Mode(
            neff=self.neff,
            cs=self.cs,
            Ex=self.Ex * other,
            Ey=self.Ey * other,
            Ez=self.Ez * other,
            Hx=self.Hx * other,
            Hy=self.Hy * other,
            Hz=self.Hz * other,
        )
        return new_mode

    __rmul__ = __mul__

    def __truediv__(self, other):
        if not isinstance(other, (float, np.floating, complex, int, np.integer)):
            raise TypeError(
                f"unsupported operand type(s) for /: 'Mode' and '{type(other).__name__}'"
            )
        return self * (1 / other)

    def __sub__(self, other):
        if not isinstance(other, Mode):
            raise TypeError(
                f"unsupported operand type(s) for -: 'Mode' and '{type(other).__name__}'"
            )
        return self + other * (-1.0)


Modes = List[Mode]


def zero_phase(mode: Mode) -> Mode:
    """normalize (zero out) the phase of a `Mode`"""
    e = np.abs(energy_density(mode))
    m, n = np.array(np.where(e == e.max()))[:, 0]
    phase = np.exp(-1j * np.angle(np.array(mode.Hx))[m][n])
    new_mode = Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=mode.Ex * phase,
        Ey=mode.Ey * phase,
        Ez=mode.Ez * phase,
        Hx=mode.Hx * phase,
        Hy=mode.Hy * phase,
        Hz=mode.Hz * phase,
    )
    if _sum_around(np.real(new_mode.Hx), m, n) < 0:
        new_mode = invert_mode(new_mode)
    return new_mode


# def _centroid_idxs(arr2d: np.ndarray) -> tuple[int, int]:
#    centroid_x = np.average(np.arange(arr2d.shape[1]), weights=arr2d.sum(axis=0))
#    centroid_y = np.average(np.arange(arr2d.shape[0]), weights=arr2d.sum(axis=1))
#    return round(float(centroid_x)), round(float(centroid_y))


def _sum_around(field, m, n, r=2):
    total = 0
    idxs = range(-r, r + 1)
    idx_tups = product(idxs, idxs)
    M, N = field.shape
    for i, j in idx_tups:
        m_ = min(m + i, M - 1) if i >= 0 else max(m + i, 0)
        n_ = min(n + j, N - 1) if j >= 0 else max(n + j, 0)
        total = total + field[m_, n_]
    return total


def invert_mode(mode: Mode) -> Mode:
    """invert a `Mode`"""
    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=-mode.Ex,
        Ey=-mode.Ey,
        Ez=-mode.Ez,
        Hx=-mode.Hx,
        Hy=-mode.Hy,
        Hz=-mode.Hz,
    )


def inner_product(mode1: Mode, mode2: Mode) -> float:
    """the inner product of a `Mode` with another `Mode` is uniquely defined."""
    mesh = mode1.mesh
    cross = mode1.Ex * mode2.Hy - mode1.Ey * mode2.Hx
    return np.trapz(np.trapz(cross, mesh.y_), mesh.x_)


def inner_product_conj(mode1: Mode, mode2: Mode) -> float:
    """the inner product of a `Mode` with another `Mode` is uniquely defined."""
    mesh = mode1.mesh
    cross = mode1.Ex * mode2.Hy.conj() - mode1.Ey * mode2.Hx.conj()
    return np.trapz(np.trapz(cross, mesh.y_), mesh.x_)


def normalize_product(mode: Mode) -> Mode:
    """normalize a `Mode` according to the `inner_product` with itself"""
    factor = np.sqrt(inner_product(mode, mode))
    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=mode.Ex / factor,
        Ey=mode.Ey / factor,
        Ez=mode.Ez / factor,
        Hx=mode.Hx / factor,
        Hy=mode.Hy / factor,
        Hz=mode.Hz / factor,
    )


def electric_energy_density(
    mode: Mode,
) -> np.ndarray[Tuple[int, int], np.dtype[np.float_]]:
    """get the electric energy density contained in a `Mode`"""
    epsx, epsy, epsz = mode.cs.nx**2, mode.cs.ny**2, mode.cs.nz**2
    return (
        0.5
        * eps0
        * (
            epsx * np.abs(mode.Ex) ** 2
            + epsy * np.abs(mode.Ey) ** 2
            + epsz * np.abs(mode.Ez) ** 2
        )
    )


def electric_energy(mode: Mode) -> float:
    """get the electric energy contained in a `Mode`"""
    return electric_energy_density(mode).sum()


def magnetic_energy_density(
    mode: Mode,
) -> np.ndarray[Tuple[int, int], np.dtype[np.float_]]:
    """get the magnetic energy density contained in a `Mode`"""
    return (
        0.5 * mu0 * (np.abs(mode.Hx) ** 2 + np.abs(mode.Hy) ** 2 + np.abs(mode.Hz) ** 2)
    )


def magnetic_energy(mode: Mode) -> float:
    """get the magnetic energy contained in a `Mode`"""
    return magnetic_energy_density(mode).sum()


def energy_density(mode: Mode) -> np.ndarray[Tuple[int, int], np.dtype[np.float_]]:
    """get the energy density contained in a `Mode`"""
    return electric_energy_density(mode) + magnetic_energy_density(mode)


def energy(mode: Mode) -> float:
    """get the energy contained in a `Mode`"""
    return energy_density(mode).sum()


def normalize_energy(mode: Mode) -> Mode:
    """normalize a mode according to the energy it contains"""
    e = np.sqrt(2 * electric_energy(mode))
    h = np.sqrt(2 * magnetic_energy(mode))

    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=mode.Ex / e,
        Ey=mode.Ey / e,
        Ez=mode.Ez / e,
        Hx=mode.Hx / h,
        Hy=mode.Hy / h,
        Hz=mode.Hz / h,
    )


def is_pml_mode(mode, threshold):
    """check whether a mode can be considered a PML mode.

    Args:
        mode: the mode to classify as PML mode or not.
        pml_mode_threshold: If the mode has more than `pml_mode_threshold` part of its
            energy in the PML, it will be removed.

    Returns:
        bool: whether the mode is a PML mode or not
    """
    threshold = min(max(threshold, 0.0), 1.0)
    if threshold > 0.999:
        return False
    numx, numy = mode.mesh.num_pml
    ed = energy_density(mode)
    m, n = ed.shape
    lft = ed[:numx, :]
    rgt = ed[m - numx :, :]
    top = ed[numx : m - numx, n:numy]
    btm = ed[numx : m - numx, n - numy :]
    rest = ed[numx : m - numx, numy : n - numy]
    pml_sum = lft.sum() + rgt.sum() + top.sum() + btm.sum()
    rest_sum = rest.sum()
    # probably propper integration considering
    # the size of the mesh cells would be better here
    is_pml = pml_sum > threshold * (rest_sum + pml_sum)
    return is_pml


def te_fraction(mode: Mode) -> float:
    """the TE polarization fraction of the `Mode`"""
    epsx = np.real(mode.cs.nx**2)
    e = np.sum(0.5 * eps0 * epsx * np.abs(mode.Ex) ** 2)
    h = np.sum(0.5 * mu0 * np.abs(mode.Hx) ** 2)
    return float(e / (e + h))


def _average(field, direction="forward", axis=0):
    direction = direction.lower()
    if not direction in ["forward", "backward"]:
        raise ValueError("direction should be 'forward' or backward")
    if not axis in [0, 1]:
        raise ValueError("axis should be zero or 1")
    elif axis == 1:
        return _average(field.T, direction=direction, axis=0).T
    average = 0.5 * (field[1:] + field[:-1])
    zero = np.zeros_like(average[:1])
    if direction == "forward":
        return np.concatenate([zero, average], axis=0)
    else:
        return np.concatenate([average, zero], axis=0)


def _interpolate_Ex(mode: Mode) -> Mode:
    # TODO: take grid spacing into account
    Ey_at_Ez = _average(mode.Ey, direction="backward", axis=1)
    Ey_at_Ex = _average(Ey_at_Ez, direction="forward", axis=0)
    Ez_at_Ex = _average(mode.Ez, direction="forward", axis=0)
    Hx_at_Hz = _average(mode.Hx, direction="forward", axis=0)
    Hx_at_Ex = _average(Hx_at_Hz, direction="backward", axis=1)
    Hz_at_Ex = _average(mode.Hz, direction="backward", axis=1)
    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=mode.Ex,
        Ey=Ey_at_Ex,
        Ez=Ez_at_Ex,
        Hx=Hx_at_Ex,
        Hy=mode.Hy,
        Hz=Hz_at_Ex,
    )


def _interpolate_Ey(mode: Mode) -> Mode:
    # TODO: take grid spacing into account
    Ex_at_Ez = _average(mode.Ex, direction="backward", axis=0)
    Ex_at_Ey = _average(Ex_at_Ez, direction="forward", axis=1)
    Ez_at_Ey = _average(mode.Ez, direction="forward", axis=1)
    Hy_at_Hz = _average(mode.Hy, direction="forward", axis=1)
    Hy_at_Ey = _average(Hy_at_Hz, direction="backward", axis=0)
    Hz_at_Ey = _average(mode.Hz, direction="backward", axis=0)
    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=Ex_at_Ey,
        Ey=mode.Ey,
        Ez=Ez_at_Ey,
        Hx=mode.Hx,
        Hy=Hy_at_Ey,
        Hz=Hz_at_Ey,
    )


def _interpolate_Ez(mode: Mode) -> Mode:
    # TODO: take grid spacing into account
    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=_average(mode.Ex, direction="backward", axis=0),
        Ey=_average(mode.Ey, direction="backward", axis=1),
        Ez=mode.Ez,
        Hx=_average(mode.Hx, direction="backward", axis=1),
        Hy=_average(mode.Hy, direction="backward", axis=0),
        Hz=_average(
            _average(mode.Hz, direction="backward", axis=0),
            direction="backward",
            axis=1,
        ),
    )


def _interpolate_Hz(mode: Mode) -> Mode:
    # TODO: take grid spacing into account
    return Mode(
        neff=mode.neff,
        cs=mode.cs,
        Ex=_average(mode.Ex, direction="forward", axis=1),
        Ey=_average(mode.Ey, direction="forward", axis=0),
        Ez=_average(
            _average(mode.Ez, direction="forward", axis=0), direction="forward", axis=1
        ),
        Hx=_average(mode.Hx, direction="forward", axis=0),
        Hy=_average(mode.Hy, direction="forward", axis=1),
        Hz=mode.Hz,
    )



"""Propagating fields throug devices"""

import jax.numpy as np
import numpy as onp
import sax
from sax.backends import circuit_backends

from meow.eme import (
    compute_interface_s_matrices,
    compute_interface_s_matrix,
    compute_propagation_s_matrices,
)
from meow.eme.sax import _validate_sax_backend


def _connect_two(l, r, sax_backend):
    """l -> left, r -> right"""
    # TODO there must be an easier way to do this...
    s_l, p_l = sax.sdense(l)
    s_r, p_r = sax.sdense(r)
    instances = {"l": l, "r": r}
    p_lr = [p for p in p_l.keys() if "right" in p]  # right ports of left
    p_rl = [p for p in p_r.keys() if "left" in p]  # left ports of right

    p_ll = [p for p in p_l.keys() if "left" in p]  # left ports of left
    p_rr = [p for p in p_r.keys() if "right" in p]  # right ports of right

    p_lr.sort()
    p_rl.sort()
    connections = {f"l,{pl}": f"r,{pr}" for pl, pr in zip(p_lr, p_rl)}
    ports = {**{p: f"l,{p}" for p in p_ll}, **{p: f"r,{p}" for p in p_rr}}
    net = dict(instances=instances, connections=connections, ports=ports)
    analyze_circuit, evaluate_circuit = circuit_backends[sax_backend]
    analyzed = analyze_circuit(net["connections"], net["ports"])
    return evaluate_circuit(analyzed, net["instances"])


def pi_pairs(propagations, interfaces, sax_backend):
    """generates the S-matrices of cells: a combination of propagation and interface matrix"""
    S = []
    for i in range(len(propagations)):
        p = propagations[f"p_{i}"]
        if i == len(interfaces):
            S.append(p)
        else:
            c = interfaces[f"i_{i}_{i+1}"]
            S.append(_connect_two(p, c, sax_backend))

    return S


def l2r_matrices(pairs, identity, sax_backend):
    Ss = [identity]

    for p in pairs[:-1]:
        Ss.append(_connect_two(Ss[-1], p, sax_backend))

    return Ss


def r2l_matrices(pairs, sax_backend):
    Ss = [pairs[-1]]

    for p in pairs[-1::-1]:
        Ss.append(_connect_two(p, Ss[-1], sax_backend))

    return Ss[::-1]


def split_square_matrix(matrix, idx):
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError("Matrix has to be square")
    return [matrix[:idx, :idx], matrix[:idx, idx:]], [
        matrix[idx:, :idx],
        matrix[idx:, idx:],
    ]


def propagate(l2rs, r2ls, excitation_l, excitation_r):
    forwards = []
    backwards = []
    for l2r, r2l in zip(l2rs, r2ls):
        s_l2r, p = sax.sdense(l2r)
        s_r2l, _ = sax.sdense(r2l)
        m = len([k for k in p.keys() if "right" in k])
        f, b = compute_mode_amplitudes(s_l2r, s_r2l, m, excitation_l, excitation_r)
        forwards.append(f)
        backwards.append(b)
    return forwards, backwards


def compute_mode_amplitudes(u, v, m, excitation_l, excitation_r):
    n = u.shape[0] - m
    l = v.shape[0] - m
    [u11, u12], [u21, u22] = split_square_matrix(u, n)
    [v11, v12], [v21, v22] = split_square_matrix(v, m)

    RHS = u21 @ excitation_l + u22 @ v12 @ excitation_r
    LHS = np.diag(np.ones(m)) - u22 @ v11
    forward = np.linalg.solve(LHS, RHS)
    backward = v12 @ excitation_r + v11 @ forward  # Attention v21 was v12
    return forward, backward


def plot_fields(modes, cells, forwards, backwards, y, z, lim=1):
    mode_set = modes[0]
    mesh_y = cells[0].mesh.y
    mesh_x = cells[0].mesh.x
    mesh_x = mesh_x[:-1] + np.diff(mesh_x) / 2
    i_y = np.argmin(np.abs(mesh_y - y))

    lim = None
    E_tot = onp.zeros((len(z), len(mesh_x)), dtype=complex)
    for mode_set, forward, backward, cell in zip(modes, forwards, backwards, cells):
        Ex = np.array(0 + 0j)
        i_min = np.argmax(z >= cell.z_min)
        i_max = np.argmax(z > cell.z_max)
        if i_max == 0:
            z_ = z[i_min:]
        else:
            z_ = z[i_min:i_max]

        z_local = z_ - cell.z_min  # [:-1] + np.diff(z_) / 2
        for mode, f, b in zip(mode_set, forward, backward):
            E_slice = mode.Ex[:, i_y]

            Ex += np.outer(
                f * E_slice.T, np.exp(2j * np.pi * mode.neff / mode.env.wl * z_local)
            )

            Ex += np.outer(
                b * E_slice.T, np.exp(-2j * np.pi * mode.neff / mode.env.wl * z_local)
            )

        if i_max == 0:
            E_tot[i_min:] = Ex.T
        else:
            E_tot[i_min:i_max] = Ex.T

        # X, Y = np.meshgrid(z_, mesh_x)
        # plt.pcolormesh(X, Y, np.abs(Ex), vmin = -lim, vmax = lim)
    # plt.xlabel("z in um")
    # plt.ylabel("x in um")
    return E_tot, mesh_x


def propagate_modes(modes, cells, ex_l, ex_r, y, z, sax_backend=None):
    sax_backend = _validate_sax_backend(sax_backend)
    propagations = compute_propagation_s_matrices(modes, cells)
    interfaces = compute_interface_s_matrices(
        modes,
        enforce_reciprocity=False,
    )
    identity = compute_interface_s_matrix(
        modes[0],
        modes[0],
        enforce_reciprocity=False,
    )

    pairs = pi_pairs(propagations, interfaces, sax_backend)
    l2rs = l2r_matrices(pairs, identity, sax_backend)
    r2ls = r2l_matrices(pairs, sax_backend)

    forwards, backwards = propagate(l2rs, r2ls, ex_l, ex_r)
    return plot_fields(modes, cells, forwards, backwards, y, z)



from mode_data import MODE_DATA

from meow import Mode

mode1 = Mode.parse_obj(MODE_DATA)
mode2 = Mode.parse_obj(MODE_DATA)


def test_multiply_modes():
    """multiplying two modes objects is not supported and should raise a TypeError"""
    try:
        mode1 * mode2
        assert False
    except TypeError:
        pass


def test_multiply_scalar():
    assert (mode1 * 3).Ex[0, 0] == 3


def test_add_modes():
    assert (mode1 + mode2).Ex[0, 0] == 2


def test_substract_modes():
    assert (mode1 - mode2).Ex[0, 0] == 0


def test_divide_scalar():
    assert (mode1 / 2).Ex[0, 0] == 1 / 2


if __name__ == "__main__":
    import pytest  # fmt: skip

    pytest.main([__file__])



""" an EME Cell """

import warnings
from typing import Dict, List, Tuple, Union, cast

import numpy as np
from pydantic.v1 import Field
from scipy.ndimage import convolve

from .base_model import BaseModel, _array, cached_property
from .materials import Material
from .mesh import Mesh2D
from .structures import (
    Structure2D,
    Structure3D,
    _classify_structures_by_mesh_order_and_material,
    _sort_structures,
)


class Cell(BaseModel):
    """A Cell defines an interval in z (the direction of propagation) within
    the simulation domain. The intersecting Structure3Ds are discretized by
    a given mesh at the center of the Cell"""

    structures: List[Structure3D] = Field(
        description="the 3D structures which will be sliced by the cell"
    )
    mesh: Mesh2D = Field(description="the mesh to discretize the structures with")
    z_min: float = Field(description="the starting z-coordinate of the cell")
    z_max: float = Field(description="the ending z-coordinate of the cell")

    @property
    def z(self):
        return 0.5 * (self.z_min + self.z_max)

    @property
    def length(self):
        return np.abs(self.z_max - self.z_min)

    @cached_property
    def materials(self):
        materials = {}
        for i, structure in enumerate(_sort_structures(self.structures), start=1):
            if not structure.material in materials:
                materials[structure.material] = i
        return materials

    @cached_property
    def structures_2d(self) -> List[Structure2D]:
        z = 0.5 * (self.z_min + self.z_max)
        list_of_list = [s._project(z) for s in self.structures]
        structures = [s for ss in list_of_list for s in ss]
        return structures

    @cached_property
    def m_full(self):
        return _create_full_material_array(
            mesh=self.mesh,
            structures=self.structures_2d,
            materials=self.materials,
        )

    def _visualize(self, ax=None, cbar=True, show=True):
        import matplotlib.pyplot as plt  # fmt: skip
        from matplotlib.colors import ListedColormap, to_rgba  # fmt: skip
        from mpl_toolkits.axes_grid1 import make_axes_locatable  # fmt: skip

        colors = [(0, 0, 0, 0)] + [
            to_rgba(m.meta.get("color", (0, 0, 0, 0))) for m in self.materials
        ]
        cmap = ListedColormap(colors=colors)  # type: ignore
        if ax is not None:
            plt.sca(ax)
        else:
            ax = plt.gca()
        plt.pcolormesh(
            self.mesh.X_full,
            self.mesh.Y_full,
            self.m_full,
            cmap=cmap,
            vmin=0,
            vmax=len(self.materials) + 1,
        )
        plt.axis("scaled")
        plt.grid(True)
        if cbar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            _cbar = plt.colorbar(
                ticks=np.concatenate(
                    [np.unique(self.m_full) + 0.5, [len(self.materials) + 1.5]]
                ),
                cax=cax,
            )
            labels = [""] + [m.name for m in self.materials] + [""]
            _cbar.ax.set_yticklabels(labels, rotation=90, va="center")
            plt.sca(ax)
        if show:
            plt.show()


def create_cells(
    structures: List[Structure3D],
    mesh: Union[Mesh2D, List[Mesh2D]],
    Ls: np.ndarray[Tuple[int], np.dtype[np.float_]],
    z_min: float = 0.0,
) -> List[Cell]:
    """easily create multiple `Cell` objects given a `Mesh` and a collection of cell lengths"""
    warnings.warn(
        "create_cells will be removed in a future version. Please create your cells in a loop instead.",
        DeprecationWarning,
    )

    Ls = np.asarray(Ls, float)
    if Ls.ndim != 1:
        raise ValueError(f"Ls should be 1D. Got shape: {Ls.shape}.")
    if Ls.shape[0] < 0:
        raise ValueError(f"length of Ls array should be nonzero. Got: {Ls}.")

    meshes = [mesh] * Ls.shape[0] if isinstance(mesh, Mesh2D) else mesh
    if len(Ls) != len(meshes):
        raise ValueError(
            f"Number of meshes should correspond to number of lengths (length of Ls). Got {len(meshes)} != {len(Ls)}."
        )

    z = np.cumsum(np.concatenate([np.asarray([z_min], float), Ls]))
    cells = [
        Cell(
            structures=structures,
            mesh=mesh,
            z_min=z_min,
            z_max=z_max,
        )
        for mesh, (z_min, z_max) in zip(meshes, zip(z[:-1], z[1:]))
    ]

    return cells


def _create_full_material_array(
    mesh: Mesh2D,
    structures: List[Structure2D],
    materials: Dict[Material, int],
):
    m_full = np.zeros_like(mesh.X_full, dtype=np.int_)
    structures_dict = _classify_structures_by_mesh_order_and_material(
        structures, materials
    )
    for structures in structures_dict.values():
        _m_full = _create_material_array(mesh, materials, structures)
        mask = _m_full > 0
        m_full[mask] = _m_full[mask]

    m_full = _fill_single_pixel_gaps(m_full)

    return m_full.view(_array)


def _create_material_array(
    mesh: Mesh2D,
    materials: Dict[Material, int],
    structures: List[Structure2D],
) -> np.ndarray:
    m_full = np.zeros_like(mesh.X_full, dtype=np.int_)
    for structure in structures:
        mask = structure.geometry._mask(mesh.X_full, mesh.Y_full)
        m_full[mask] = materials[structure.material]

    if mesh.ez_interfaces:
        mask_ez_horizontal = np.zeros_like(m_full, dtype=bool)
        mask_ez_horizontal[:, ::2] = True
        mask_ez_vertical = np.zeros_like(m_full, dtype=bool)
        mask_ez_vertical[::2, :] = True
        mask_boundaries_vertical = _get_boundary_mask_vertical(m_full)
        mask_boundaries_vertical = mask_boundaries_vertical & (~mask_ez_vertical)
        mask_boundaries_horizontal = _get_boundary_mask_horizontal(m_full)
        mask_boundaries_horizontal = mask_boundaries_horizontal & (~mask_ez_horizontal)
        mask_temp = mask_boundaries_vertical | mask_boundaries_horizontal
        mask_corner_left = _fill_corner_left_mask(mask_temp)
        mask_corner_right = _fill_corner_right_mask(mask_temp)
        mask_to_remove = mask_temp | mask_corner_left | mask_corner_right
        mask_to_keep = (m_full > 0) & (~mask_to_remove)

        mask_ez = np.zeros_like(m_full, dtype=bool)
        mask_ez[::2, ::2] = True
        final_mask_to_keep = mask_to_keep & mask_ez

        mask_ex = np.zeros_like(m_full, dtype=bool)
        mask_ex[1::2, ::2] = True
        final_mask_to_keep |= mask_to_keep & mask_ex

        mask_ey = np.zeros_like(m_full, dtype=bool)
        mask_ey[::2, 1::2] = True
        final_mask_to_keep |= mask_to_keep & mask_ey

        mask_hz = np.zeros_like(m_full, dtype=bool)
        mask_hz[1::2, 1::2] = True
        final_mask_to_keep |= mask_to_keep & mask_hz

        m_full[~final_mask_to_keep] = 0
    return m_full


def _fill_corner_left_mask(mask):
    return (
        convolve(np.asarray(mask, dtype=float), np.array([[-1.0, +1.0], [+1.0, -1.0]]))
        > 1.0
    )  # type: ignore


def _fill_corner_right_mask(mask):
    return (
        convolve(
            np.asarray(mask, dtype=float),
            np.array([[0.0, 0.0], [+1.0, -1.0], [-1.0, +1.0]]),
        )
        > 1
    )  # type: ignore


def _get_boundary_mask_horizontal(m_full, negate=False):
    mask = np.zeros((m_full.shape[0] + 2, m_full.shape[1] + 2), dtype=bool)
    mask[1:-1, 1:-1] = m_full > 0
    if negate:
        mask = ~mask
    f = np.ndarray[Tuple[int, int], np.dtype[np.float_]]
    conv1 = cast(f, convolve(np.array(mask[:, :], dtype=int), np.array([[-1, 1]])))
    conv2 = cast(f, convolve(np.array(mask[:, ::-1], dtype=int), np.array([[-1, 1]])))
    conv3 = cast(f, convolve(np.array(mask[::-1, :], dtype=int), np.array([[-1, 1]])))
    mask1 = (conv1 > 0.0)[:, :]
    mask2 = (conv2 > 0.0)[:, ::-1]
    mask3 = (conv3 > 0.0)[::-1, :]
    mask = (mask1 | mask2 | mask3)[1:-1, 1:-1]
    # don't mask edge of simulation area:
    mask[:, 0] = mask[:, -1] = False
    return mask


def _get_boundary_mask_vertical(m_full, negate=False):
    return _get_boundary_mask_horizontal(m_full.T, negate=negate).T


def _fill_single_pixel_gaps(m_full):
    mat_x = np.maximum(
        np.roll(m_full, shift=1, axis=0),
        np.roll(m_full, shift=-1, axis=0),
    )
    mat_y = np.maximum(
        np.roll(m_full, shift=1, axis=1),
        np.roll(m_full, shift=-1, axis=1),
    )
    mat = mat_x | mat_y

    mask = m_full > 0
    mask_x = np.roll(mask, shift=1, axis=0) & (~mask) & np.roll(mask, shift=-1, axis=0)
    mask_y = np.roll(mask, shift=1, axis=1) & (~mask) & np.roll(mask, shift=-1, axis=1)
    mask = mask_x | mask_y
    mask[:1, :] = False
    mask[-1:, :] = False
    mask[:, :1] = False
    mask[:, -1:] = False
    m_full[mask] = mat[mask]
    return m_full



""" SAX backend for EME (default backend) """

from typing import List, Optional

import numpy as np
import sax
from sax.backends import circuit_backends
from sax.utils import get_ports

from ..base_model import _array
from ..cell import Cell
from ..mode import Mode
from .common import (
    DEFAULT_CONJUGATE,
    DEFAULT_ENFORCE_LOSSY_UNITARITY,
    DEFAULT_ENFORCE_RECIPROCITY,
    compute_interface_s_matrices,
    compute_propagation_s_matrices,
)

try:
    import klujax


except ImportError:
    klujax = None


def _get_netlist(propagations, interfaces):
    """get the netlist of a stack of `Modes`"""

    instances = {
        **{k: S for k, S in propagations.items()},
        **{k: S for k, S in interfaces.items()},
    }

    propagation_keys = list(propagations)
    connections = {}
    for i in range(len(interfaces)):
        for port_mode in get_ports(interfaces[f"i_{i}_{i+1}"]):
            other_port_mode = _other_port(port_mode)
            if "left" in port_mode:
                connections[f"p_{i},{other_port_mode}"] = f"i_{i}_{i+1},{port_mode}"
            elif "right" in port_mode:
                connections[f"i_{i}_{i+1},{port_mode}"] = f"p_{i+1},{other_port_mode}"

    ports = {}
    for port_mode in get_ports(propagations[propagation_keys[0]]):
        if "right" in port_mode:
            continue
        ports[port_mode] = f"{propagation_keys[0]},{port_mode}"
    for port_mode in get_ports(propagations[propagation_keys[-1]]):
        if "left" in port_mode:
            continue
        ports[port_mode] = f"{propagation_keys[-1]},{port_mode}"

    net = {"instances": instances, "connections": connections, "ports": ports}
    return net


def _other_port(port_mode):
    if "left" in port_mode:
        return port_mode.replace("left", "right")
    elif "right" in port_mode:
        return port_mode.replace("right", "left")
    else:
        return port_mode


def _validate_sax_backend(sax_backend):
    if sax_backend is None:
        sax_backend = "klu" if klujax is not None else "default"

    if sax_backend not in ["default", "klu"]:
        raise ValueError(
            f"Invalid SAX Backend. Got: {sax_backend!r}. Should be 'default' or 'klu'."
        )
    return sax_backend


def compute_s_matrix_sax(
    modes: List[List[Mode]],
    cells: Optional[List[Cell]] = None,
    cell_lengths: Optional[List[float]] = None,
    sax_backend: Optional[str] = None,
    conjugate: bool = DEFAULT_CONJUGATE,
    enforce_reciprocity: bool = DEFAULT_ENFORCE_RECIPROCITY,
    enforce_lossy_unitarity: bool = DEFAULT_ENFORCE_LOSSY_UNITARITY,
    **kwargs,
):
    """Calculate the S-matrix for given sets of modes, each set belonging to a `Cell`

    Args:
        modes: Each collection of modes for each of the `Cell` objects
        backend: which SAX backend to use to calculate the final S-matrix.
    """
    sax_backend = _validate_sax_backend(sax_backend)
    _compute_propagation_s_matrices = kwargs.pop(
        "compute_propagation_s_matrices", compute_propagation_s_matrices
    )
    _compute_interface_s_matrices = kwargs.pop(
        "compute_interface_s_matrices", compute_interface_s_matrices
    )
    propagations = _compute_propagation_s_matrices(
        modes, cells, cell_lengths=cell_lengths
    )
    interfaces = _compute_interface_s_matrices(
        modes,
        conjugate=conjugate,
        enforce_reciprocity=enforce_reciprocity,
        enforce_lossy_unitarity=enforce_lossy_unitarity,
        **kwargs,
    )

    # TODO: fix SAX Multimode to reduce this ad-hoc SAX-hacking.
    net = _get_netlist(propagations, interfaces)
    analyze_circuit, evaluate_circuit = circuit_backends[sax_backend]
    analyzed = analyze_circuit(net["connections"], net["ports"])
    S, port_map = sax.sdense(
        evaluate_circuit(
            analyzed,
            instances=net["instances"],
        )
    )
    S = np.asarray(S).view(_array)

    # final sorting of result:
    current_port_map = {
        (p, int(i)): j
        for (p, i), j in {
            tuple(pm.split("@")): idx for pm, idx in port_map.items()
        }.items()
    }
    desired_port_map = {pm: i for i, pm in enumerate(sorted(current_port_map))}
    idxs = [current_port_map[pm] for pm in desired_port_map]
    S = S[idxs, :][:, idxs]
    port_map = {f"{p}@{m}": v for (p, m), v in desired_port_map.items()}

    return S, port_map



""" a 2D Mesh """

import warnings
from functools import wraps
from typing import Literal, Optional, Tuple

import numpy as np
from pydantic.v1 import Field
from pydantic.v1.types import NonNegativeInt

from .base_model import BaseModel, _array, cached_property


class Mesh(BaseModel):
    """[BaseClass] a ``Mesh`` describes how a ``Structure3D`` is discretized"""


class Mesh2D(Mesh):
    """a 2D Mesh or ``Mesh2D`` describes how a ``Structure3D`` is discritized into a ``Cell`` or ``CrossSection``"""

    x: np.ndarray[Tuple[int], np.dtype[np.float_]] = Field(
        description="x-coordinates of the mesh (Ez locations, i.e. corners of the 2D cell)"
    )
    y: np.ndarray[Tuple[int], np.dtype[np.float_]] = Field(
        description="y-coordinates of the mesh (Ez locations, i.e. corners of the 2D cell)"
    )
    angle_phi: float = Field(
        default=0.0,
        description="Azimuth angle of the propagation axis in the plane orthogonal to the mesh.",
    )
    angle_theta: float = Field(
        default=0.0,
        description="Polar angle of the propagation axis from the injection axis.",
    )
    bend_radius: Optional[float] = Field(
        default=None,
        description=(
            "A curvature radius for simulation of waveguide bends. "
            "Tidy3D: Can be negative, in which case the mode plane center has a smaller value than "
            "the curvature center along the tangential axis perpendicular to the bend axis."
        ),
    )
    bend_axis: Optional[Literal[0, 1]] = Field(
        default=None,
        description=(
            "Index into the two tangential axes defining the normal to the plane in which the bend lies. "
            "This must be provided if ``bend_radius`` is not ``None``. For example, for a ring in the "
            "global xy-plane, and a mode plane in either the xz or the yz plane, the ``bend_axis`` is "
            "always 1 (the global z axis)."
        ),
    )
    num_pml: Tuple[NonNegativeInt, NonNegativeInt] = Field(
        default=(0, 0),
        description="Number of standard pml layers to add in the two tangential axes.",
    )

    ez_interfaces: bool = Field(
        default=False,
        description=(
            "when enabled, the meshing algorithm will throw away any index values "
            "at the interfaces which are not on even (Ez) half-grid locations. "
            "Enabling this should result in more symmetric modes."
        ),
    )

    @cached_property
    def dx(self):
        """dx at Hz locations, i.e. center of the 2D cell"""
        return (self.x[1:] - self.x[:-1]).view(_array)

    @cached_property
    def dy(self):
        """dy at Hz locations, i.e. center of the 2D cell"""
        return (self.y[1:] - self.y[:-1]).view(_array)

    @cached_property
    def x_(self):
        """x at Hz locations, i.e. center of the 2D cell"""
        return 0.5 * (self.x[1:] + self.x[:-1]).view(_array)

    @cached_property
    def y_(self):
        """y at Hz locations, i.e. center of the 2D cell"""
        return 0.5 * (self.y[1:] + self.y[:-1]).view(_array)

    @cached_property
    def x_full(self):
        """x at half-integer locations"""
        return (
            np.stack([self.x[:-1], self.x[:-1] + self.dx / 2], 1).ravel().view(_array)
        )

    @cached_property
    def y_full(self):
        """y at half-integer locations"""
        return (
            np.stack([self.y[:-1], self.y[:-1] + self.dy / 2], 1).ravel().view(_array)
        )

    @cached_property
    def XY_full(self):
        """X and Y at half-integer locations"""
        Y_full, X_full = np.meshgrid(self.y_full, self.x_full)
        return X_full.view(_array), Y_full.view(_array)

    @property
    def X_full(self):
        """X at half-integer locations"""
        return self.XY_full[0].view(_array)

    @property
    def Y_full(self):
        """Y at half-integer locations"""
        return self.XY_full[1].view(_array)

    @property
    def Xx(self):
        """X at Ex locations"""
        return self.X_full[1::2, ::2].view(_array)

    @property
    def Yx(self):
        """Y at Ex locations"""
        return self.Y_full[1::2, ::2].view(_array)

    @property
    def Xy(self):
        """X at Ey locations"""
        return self.X_full[::2, 1::2].view(_array)

    @property
    def Yy(self):
        """Y at Ey locations"""
        return self.Y_full[::2, 1::2].view(_array)

    @property
    def Xz(self):
        """X at Ez locations"""
        return self.X_full[::2, ::2].view(_array)

    @property
    def Yz(self):
        """Y at Ez locations"""
        return self.Y_full[::2, ::2].view(_array)

    @property
    def Xz_(self):
        """X at Hz locations"""
        return self.X_full[1::2, 1::2].view(_array)

    @property
    def Yz_(self):
        """Y at Hz locations"""
        return self.Y_full[1::2, 1::2].view(_array)

    def __eq__(self, other):
        eq = True
        for k, v in self.dict().items():
            if isinstance(v, np.ndarray):
                eq &= bool(((v - getattr(other, k)) < 1e-6).all())
            else:
                eq &= bool(v == getattr(other, k))
        return eq


@wraps(Mesh2D)
def Mesh2d(*args, **kwargs):
    warnings.warn(
        "Mesh2d is deprecated. Please use Mesh2D (with capital D in the end).",
        DeprecationWarning,
    )
    return Mesh2D(*args, **kwargs)



import numpy as np
from mode_data import MODE_DATA

import meow as mw


def test_deserialization():
    simple_mode = mw.Mode.parse_obj(MODE_DATA)
    assert np.round(np.real(simple_mode.neff), 6) == np.round(
        MODE_DATA["neff"]["real"], 6
    )
    assert np.round(np.imag(simple_mode.neff), 6) == np.round(
        MODE_DATA["neff"]["imag"], 6
    )



""" MEOW: Modeling of Eigenmodes and Overlaps in Waveguides """

__author__ = "Floris Laporte"
__version__ = "0.8.0"

import warnings

# Silence Excessive Logging...

try:
    from loguru import logger

    logger.disable("gdsfactory")
except ImportError:
    pass

try:
    from numexpr.utils import log

    log.setLevel("CRITICAL")
except ImportError:
    pass

try:
    from rich import pretty

    old_install = pretty.install
    pretty.install = lambda *_, **__: None
    import tidy3d

    pretty.install = old_install
except ImportError:
    pass


try:
    import sax

    warnings.filterwarnings(action="ignore", module="sax")
except ImportError:
    pass

from . import base_model as base_model
from . import cell as cell
from . import cross_section as cross_section
from . import eme as eme
from . import environment as environment
from . import fde as fde
from . import gds_structures as gds_structures
from . import geometries as geometries
from . import materials as materials
from . import mesh as mesh
from . import mode as mode
from . import structures as structures

# from . import visualize as visualize
from .base_model import BaseModel as BaseModel
from .cell import Cell as Cell
from .cell import create_cells as create_cells
from .cross_section import CrossSection as CrossSection
from .eme import compute_interface_s_matrices as compute_interface_s_matrices
from .eme import compute_interface_s_matrix as compute_interface_s_matrix
from .eme import compute_propagation_s_matrices as compute_propagation_s_matrices
from .eme import compute_propagation_s_matrix as compute_propagation_s_matrix
from .eme import compute_s_matrix as compute_s_matrix
from .eme import compute_s_matrix_sax as compute_s_matrix_sax
from .eme import select_ports as select_ports
from .environment import Environment as Environment
from .fde import compute_modes as compute_modes
from .fde import compute_modes_lumerical as compute_modes_lumerical
from .fde import compute_modes_tidy3d as compute_modes_tidy3d
from .gds_structures import GdsExtrusionRule as GdsExtrusionRule
from .gds_structures import extrude_gds as extrude_gds
from .geometries import Box as Box
from .geometries import Geometry2D as Geometry2D
from .geometries import Geometry3D as Geometry3D
from .geometries import Prism as Prism
from .geometries import Rectangle as Rectangle
from .materials import Material as Material
from .materials import SampledMaterial as SampledMaterial
from .materials import TidyMaterial as TidyMaterial
from .materials import silicon as silicon
from .materials import silicon_nitride as silicon_nitride
from .materials import silicon_oxide as silicon_oxide
from .mesh import Mesh as Mesh
from .mesh import Mesh2D as Mesh2D
from .mesh import Mesh2d as Mesh2d
from .mode import Mode as Mode
from .mode import electric_energy as electric_energy
from .mode import electric_energy_density as electric_energy_density
from .mode import energy as energy
from .mode import energy_density as energy_density
from .mode import inner_product as inner_product
from .mode import inner_product_conj as inner_product_conj
from .mode import invert_mode as invert_mode
from .mode import is_pml_mode as is_pml_mode
from .mode import magnetic_energy as magnetic_energy
from .mode import magnetic_energy_density as magnetic_energy_density
from .mode import normalize_energy as normalize_energy
from .mode import normalize_product as normalize_product
from .mode import te_fraction as te_fraction
from .mode import zero_phase as zero_phase
from .structures import Structure as Structure
from .structures import Structure2D as Structure2D
from .structures import Structure3D as Structure3D
from .visualize import vis as vis
from .visualize import visualize as visualize



""" Visualizations for common meow-datatypes """

from collections.abc import Iterable
from typing import Any, Callable

import numpy as np

from meow.structures import _visualize_structures

try:
    import matplotlib.pyplot as plt  # fmt: skip


except ImportError:
    plt = None

try:
    import gdsfactory as gf  # fmt: skip
except ImportError:
    gf = None

try:
    from jaxlib.xla_extension import DeviceArray  # fmt: skip # type: ignore
except ImportError:
    DeviceArray = None


def _visualize_s_matrix(S, fmt=None, title=None, show=True, phase=False, ax=None):
    import matplotlib.pyplot as plt  # fmt: skip
    if phase:
        fmt = ".0f"
    else:
        fmt = ".3f"

    Z = np.abs(S)
    y, x = np.arange(Z.shape[0])[::-1], np.arange(Z.shape[1])
    Y, X = np.meshgrid(y, x)

    if ax:
        plt.sca(ax)
    else:
        plt.figure(figsize=(2 * x.shape[0] / 3, 2 * y.shape[0] / 3))

    plt.pcolormesh(X, Y, Z[::-1].T, cmap="Greys", vmin=0.0, vmax=2.0 * Z.max())

    coords_ = np.concatenate(
        [np.array([x[0] - 1]), x, np.array([x[-1] + 1])], axis=0, dtype=float
    )
    labels = ["" for _ in coords_]
    plt.xticks(coords_ + 0.5, labels)
    plt.xlim(coords_[0] + 0.5, coords_[-1] - 0.5)

    coords_ = np.concatenate(
        [np.array([y[0] + 1]), y, np.array([y[-1] - 1])], axis=0, dtype=float
    )
    coords_ = coords_[::-1]  # reverse
    labels = ["" for _ in coords_]
    plt.yticks(coords_ + 0.5, labels)
    plt.ylim(coords_[-1] - 0.5, coords_[0] + 0.5)
    plt.grid(True)

    for x, y, z in zip(X.ravel(), Y.ravel(), S[::-1].T.ravel()):
        if np.abs(z) > 0.0005:
            if phase:
                z = np.angle(z) * 180 / np.pi
            text = eval(f"f'{{z:{fmt}}}'")  # 😅
            text = text.replace("+", "\n+")
            text = text.replace("-", "\n-")
            if text[0] == "\n":
                text = text[1:]
            plt.text(x, y, text, ha="center", va="center", fontsize=8)

    if title is not None:
        plt.title(title)

    if show:
        plt.show()


def _visualize_s_pm_matrix(Spm, fmt=None, title=None, show=True, phase=False, ax=None):
    import matplotlib.pyplot as plt  # fmt: skip

    S, pm = Spm
    _visualize_s_matrix(S, fmt=fmt, title=title, show=False, phase=phase, ax=ax)
    num_left = len([p for p in pm if "left" in p])
    Z = np.abs(S)
    _, x = np.arange(Z.shape[0])[::-1], np.arange(Z.shape[1])

    plt.axvline(x[num_left] - 0.5, color="red")
    plt.axhline(x[num_left] - 0.5, color="red")

    if show:
        plt.show()


def _visualize_overlap_density(
    two_modes,
    conjugated=True,
    x_symmetry=False,
    y_symmetry=False,
    ax=None,
    n_cmap=None,
    mode_cmap=None,
    num_levels=8,
    show=True,
):
    import matplotlib.pyplot as plt  # fmt: skip

    from .mode import Mode  # fmt: skip

    mode1, mode2 = two_modes
    if conjugated:
        cross = mode1.Ex * mode2.Hy.conj() - mode1.Ey * mode2.Hx.conj()
    else:
        cross = mode1.Ex * mode2.Hy - mode1.Ey * mode2.Hx
    if x_symmetry:
        cross = 0.5 * (cross + cross[::-1])
    if y_symmetry:
        cross = 0.5 * (cross + cross[:, ::-1])
    zeros = np.zeros_like(cross)
    overlap = Mode(
        neff=mode1.neff,
        cs=mode1.cs,
        Ex=cross,
        Ey=zeros,
        Ez=zeros,
        Hx=zeros,
        Hy=zeros,
        Hz=zeros,
    )
    if ax is None:
        W, H = _figsize_visualize_mode(mode1.cs, 5)
        _, ax = plt.subplots(1, 3, figsize=(3 * W, H))

    field = "Ex" if mode1.te_fraction > 0.5 else "Hx"
    mode1._visualize(
        title=f"mode 1: {field}",
        fields=[field],
        ax=ax[0],
        n_cmap=n_cmap,
        mode_cmap=mode_cmap,
        num_levels=num_levels,
        show=False,
    )

    field = "Ex" if mode2.te_fraction > 0.5 else "Hx"
    mode2._visualize(
        title=f"mode 2: {field}",
        fields=[field],
        ax=ax[1],
        n_cmap=n_cmap,
        mode_cmap=mode_cmap,
        num_levels=num_levels,
        show=False,
    )

    title = "overlap density" + ("" if conjugated else " (no conjugations)")
    p = overlap._visualize(
        title=title,
        fields=["Ex"],
        ax=ax[2],
        n_cmap=n_cmap,
        mode_cmap=mode_cmap,
        num_levels=num_levels,
        show=False,
    )

    if show:
        plt.show()

    return p


def _visualize_gf_component(comp):
    import gdsfactory as gf  # fmt: skip

    gf.plot(comp)  # type: ignore


def _is_two_tuple(obj):
    if not isinstance(obj, tuple):
        return False
    try:
        x, y = obj  # type: ignore
        return True
    except Exception:
        return False


def _figsize_visualize_mode(cs, W0):
    x_min, x_max = cs.mesh.x.min(), cs.mesh.x.max()
    y_min, y_max = cs.mesh.y.min(), cs.mesh.y.max()
    delta_x = x_max - x_min
    delta_y = y_max - y_min
    aspect = delta_y / delta_x
    W, H = W0 + 1, W0 * aspect + 1
    return W, H


def _visualize_modes(
    modes,
    n_cmap=None,
    mode_cmap=None,
    num_levels=8,
    operation=lambda x: np.abs(x) ** 2,
    show=True,
    plot_width=6.4,
    fields=("Ex", "Hx"),
    ax=None,
):
    import matplotlib.pyplot as plt  # fmt: skip

    num_modes = len(modes)
    cs = modes[0].cs
    W, H = _figsize_visualize_mode(cs, plot_width)

    if ax is None:
        fig, ax = plt.subplots(
            num_modes,
            2,
            figsize=(2 * W, num_modes * H),
            sharex=True,
            sharey=True,
            squeeze=False,
        )
    else:
        fig = None
    for i, m in enumerate(modes):
        m._visualize(
            title=None,
            title_prefix=f"m{i}: ",
            fields=list(fields),
            ax=ax[i],
            n_cmap=n_cmap,
            mode_cmap=mode_cmap,
            num_levels=num_levels,
            operation=operation,
            show=False,
        )
    if fig is not None:
        fig.subplots_adjust(hspace=0, wspace=2 / (2 * W))
    if show:
        plt.show()


def _visualize_base_model(obj, **kwargs):
    return obj._visualize(**kwargs)


def _is_mode_list(obj: Any) -> bool:
    from .mode import Mode  # fmt: skip
    return isinstance(obj, Iterable) and all(isinstance(o, Mode) for o in obj)


def _is_structure_3d_list(obj: Any) -> bool:
    from .structures import Structure3D  # fmt: skip
    return isinstance(obj, Iterable) and all(isinstance(o, Structure3D) for o in obj)


def _is_base_model(obj: Any) -> bool:
    from .base_model import BaseModel  # fmt: skip
    return isinstance(obj, BaseModel)


def _is_mode_overlap(obj: Any) -> bool:
    from .mode import Mode  # fmt: skip
    return _is_two_tuple(obj) and all(isinstance(o, Mode) for o in obj)


def _is_s_matrix(obj: Any):
    return (
        (
            isinstance(obj, np.ndarray)
            or (DeviceArray is not None and isinstance(obj, DeviceArray))
        )
        and obj.ndim == 2
        and obj.shape[0] > 1
        and obj.shape[1] > 1
    )


def _is_s_pm_matrix(obj):
    return _is_two_tuple(obj) and _is_s_matrix(obj[0]) and isinstance(obj[1], dict)


def _is_gf_component(obj):
    return gf is not None and isinstance(obj, gf.Component)


VISUALIZATION_MAPPING: dict[Callable, Callable] = {
    _is_base_model: _visualize_base_model,
    _is_mode_list: _visualize_modes,
    _is_structure_3d_list: _visualize_structures,
    _is_mode_overlap: _visualize_overlap_density,
    _is_s_matrix: _visualize_s_matrix,
    _is_s_pm_matrix: _visualize_s_pm_matrix,
    _is_gf_component: _visualize_gf_component,
}


def visualize(obj: Any, **kwargs: Any):
    """visualize any meow object

    Args:
        obj: the meow object to visualize
        **kwargs: extra configuration to visualize the object

    Note:
        Most meow objects have a `._visualize` method.
        Check out its help to see which kwargs are accepted.
    """
    if plt is None:
        print(obj)
        return

    for check_func, vis_func in VISUALIZATION_MAPPING.items():
        if check_func(obj):
            return vis_func(obj, **kwargs)
    print(obj)


vis = visualize  # shorthand for visualize



""" FDE Tidy3d backend (default backend for MEOW) """

from types import SimpleNamespace
from typing import Literal, Optional

import numpy as np
import tidy3d
from packaging import version
from pydantic.v1 import validate_arguments
from pydantic.v1.types import PositiveFloat, PositiveInt
from scipy.constants import c
from tidy3d.plugins.mode.solver import compute_modes as _compute_modes

from ..cross_section import CrossSection
from ..mode import Mode, Modes, is_pml_mode, normalize_product, zero_phase


@validate_arguments
def compute_modes_tidy3d(
    cs: CrossSection,
    num_modes: PositiveInt = 10,
    target_neff: Optional[PositiveFloat] = None,
    precision: Literal["single", "double"] = "double",
    pml_mode_threshold: float = 1.0,
) -> Modes:
    """compute ``Modes`` for a given ``CrossSection``

    Args:
        cs: The ``CrossSection`` to calculate the modes for
        num_modes: Number of modes returned by mode solver.
        target_neff: Guess for initial effective index of the mode.
        pml_mode_threshold: If the mode has more than `pml_mode_threshold` part of its
            energy in the PML, it will be removed.
            default: 1.0 = 100% = no fitering.
    """

    if num_modes < 1:
        raise ValueError("You need to request at least 1 mode.")

    od = np.zeros_like(cs.nx)  # off diagonal entry
    new_tidy3d = version.parse(tidy3d.__version__) >= version.parse("2.2.0")
    if new_tidy3d:
        eps_cross = [cs.nx**2, od, od, od, cs.ny**2, od, od, od, cs.nz**2]
    else:
        eps_cross = [cs.nx**2, cs.ny**2, cs.nz**2]

    mode_spec = SimpleNamespace(  # tidy3d.ModeSpec alternative (prevents type checking)
        num_modes=num_modes,
        target_neff=target_neff,
        num_pml=cs.mesh.num_pml,
        filter_pol=None,
        angle_theta=cs.mesh.angle_theta,
        angle_phi=cs.mesh.angle_phi,
        bend_radius=cs.mesh.bend_radius,
        precision=precision,
        bend_axis=cs.mesh.bend_axis,
        track_freq="central",
        group_index_step=False,
    )

    ((Ex, Ey, Ez), (Hx, Hy, Hz)), neffs = (
        x.squeeze()
        for x in _compute_modes(
            eps_cross=eps_cross,
            coords=[cs.mesh.x, cs.mesh.y],
            freq=c / (cs.env.wl * 1e-6),
            mode_spec=mode_spec,
        )[:2]
    )

    if num_modes == 1:
        modes = [
            Mode(
                cs=cs,
                Ex=Ex,
                Ey=Ey,
                Ez=Ez,
                Hx=Hx,
                Hy=Hy,
                Hz=Hz,
                neff=float(neffs.real) + 1j * float(neffs.imag),
            )
            for _ in range(num_modes)
        ]
    else:  # num_modes > 1
        modes = [
            Mode(
                cs=cs,
                Ex=Ex[..., i],
                Ey=Ey[..., i],
                Ez=Ez[..., i],
                Hx=Hx[..., i],
                Hy=Hy[..., i],
                Hz=Hz[..., i],
                neff=neffs[i],
            )
            for i in range(num_modes)
        ]

    modes = [zero_phase(normalize_product(mode)) for mode in modes]
    modes = sorted(modes, key=lambda m: float(np.real(m.neff)), reverse=True)
    modes = [m for m in modes if not is_pml_mode(m, pml_mode_threshold)]

    return modes



import warnings
from secrets import token_hex
from typing import Dict, List, Literal, Tuple, Union, cast

import matplotlib.pyplot as plt
import numpy as np
import shapely.geometry as sg
from matplotlib.patches import Rectangle as MplRect
from pydantic.v1 import Field, validator

from meow.base_model import BaseModel

GEOMETRIES_3D: Dict[str, type] = {}
GEOMETRIES_2D: Dict[str, type] = {}
AxisDirection = Union[Literal["x"], Literal["y"], Literal["z"]]


class Geometry2D(BaseModel):
    type: str = ""

    def __init_subclass__(cls):
        GEOMETRIES_2D[cls.__name__] = cls

    def __new__(cls, **kwargs):
        cls = GEOMETRIES_2D.get(kwargs.get("type", cls.__name__), cls)
        return BaseModel.__new__(cls)  # type: ignore

    @validator("type", pre=True, always=True)
    def validate_type(cls, value):
        if not value:
            value = getattr(cls, "__name__", "Geometry")
        if value not in GEOMETRIES_2D:
            raise ValueError(
                f"Invalid Geometry type. Got: {value!r}. Valid types: {GEOMETRIES_2D}."
            )
        return value

    def _mask(self, X, Y):
        raise NotImplementedError(f"{self.__class__.__name__!r} cannot be masked.")

    def _visualize(self, color=None):
        raise NotImplementedError(f"{self.__class__.__name__!r} cannot be visualized.")


class Rectangle(Geometry2D):
    """a Rectangle"""

    x_min: float = Field(description="the minimum x-value of the box")
    x_max: float = Field(description="the maximum x-value of the box")
    y_min: float = Field(description="the minimum y-value of the box")
    y_max: float = Field(description="the maximum y-value of the box")

    def _mask(self, X, Y):
        mask = (
            (self.x_min <= X)
            & (X <= self.x_max)
            & (self.y_min <= Y)
            & (Y <= self.y_max)
        )
        return mask

    def _visualize(self, ax=None, show=True, color=None):
        if ax is None:
            ax = plt.gca()
        if color is None:
            color = "grey"
        width = self.x_max - self.x_min
        height = self.y_max - self.y_min
        mpl_rect = MplRect(
            xy=(self.x_min, self.y_min),
            width=width,
            height=height,
            color=color,
        )
        ax.add_patch(mpl_rect)
        ax.set_xlim(self.x_min - 0.1 * width, self.x_max + 0.1 * width)
        ax.set_ylim(self.y_min - 0.1 * width, self.y_max + 0.1 * width)
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        plt.grid(True)
        if show:
            plt.show()


class Geometry3D(BaseModel):
    type: str = ""

    def __init_subclass__(cls):
        GEOMETRIES_3D[cls.__name__] = cls

    def __new__(cls, **kwargs):
        cls = GEOMETRIES_3D.get(kwargs.get("type", cls.__name__), cls)
        return BaseModel.__new__(cls)  # type: ignore

    @validator("type", pre=True, always=True)
    def validate_type(cls, value):
        if not value:
            value = getattr(cls, "__name__", "Geometry")
        if value not in GEOMETRIES_3D:
            raise ValueError(
                f"Invalid Geometry type. Got: {value!r}. Valid types: {GEOMETRIES_3D}."
            )
        return value

    def _project(self, z: float) -> List[Geometry2D]:
        raise NotImplementedError(f"{self.__class__.__name__!r} cannot be projected.")

    def _visualize(self, scale=None):
        from trimesh.scene import Scene  # fmt: skip
        from trimesh.transformations import rotation_matrix  # fmt: skip

        scene = Scene()
        scene.add_geometry(self._trimesh(scale=scale))
        scene.apply_transform(rotation_matrix(-np.pi / 6, (0, 1, 0)))
        return scene.show()

    def _lumadd(self, sim, material_name, mesh_order, unit=1e-6, xyz="yzx"):
        raise NotImplementedError(
            f"{self.__class__.__name__!r} cannot be added to Lumerical."
        )

    def _trimesh(self, **kwargs):
        raise NotImplementedError(f"{self.__class__.__name__!r} cannot be visualized.")


class Box(Geometry3D):
    """A Box is a simple rectangular cuboid"""

    x_min: float = Field(description="the minimum x-value of the box")
    x_max: float = Field(description="the maximum x-value of the box")
    y_min: float = Field(description="the minimum y-value of the box")
    y_max: float = Field(description="the maximum y-value of the box")
    z_min: float = Field(description="the minimum z-value of the box")
    z_max: float = Field(description="the maximum z-value of the box")

    def _project(self, z: float) -> List[Geometry2D]:
        if z < self.z_min or z > self.z_max:
            return []
        rect = Rectangle(
            x_min=self.x_min,
            x_max=self.x_max,
            y_min=self.y_min,
            y_max=self.y_max,
        )
        return [rect]

    def _lumadd(self, sim, material_name, mesh_order, unit, xyz):
        x, y, z = xyz
        name = token_hex(4)
        kwargs = {
            f"{x}_min": float(self.x_min * unit),
            f"{x}_max": float(self.x_max * unit),
            f"{y}_min": float(self.y_min * unit),
            f"{y}_max": float(self.y_max * unit),
            f"{z}_min": float(self.z_min * unit),
            f"{z}_max": float(self.z_max * unit),
        }
        sim.addrect(
            name=name,
            material=material_name,
            override_mesh_order_from_material_database=True,
            mesh_order=mesh_order,
            **kwargs,
        )

    def _trimesh(self, color=None, scale=None):
        from trimesh import Trimesh  # fmt: skip
        from trimesh.creation import extrude_polygon  # fmt: skip

        sx, sy, sz = scale or (1, 1, 1)
        poly = sg.Polygon(
            [
                (self.x_min * sx, self.y_min * sy),
                (self.x_min * sx, self.y_max * sy),
                (self.x_max * sx, self.y_max * sy),
                (self.x_max * sx, self.y_min * sy),
            ],
        )
        prism = extrude_polygon(poly, self.z_max * sz - self.z_min * sz)
        prism = cast(Trimesh, prism.apply_translation((0, 0, self.z_min * sz)))
        if color is not None:
            prism.visual.face_colors = _to_rgba(color)  # type: ignore
        return prism


class Prism(Geometry3D):
    """A prism is a 2D Polygon extruded along a certain axis direction ('x', 'y', or 'z')."""

    poly: np.ndarray[Tuple[int, Literal[2]], np.dtype[np.float_]] = Field(
        description="the 2D array (Nx2) with polygon vertices"
    )
    h_min: float = Field(description="the start height of the extrusion")
    h_max: float = Field(description="the end height of the extrusion")
    axis: AxisDirection = Field(
        default="y",
        description="the axis along which the polygon will be extruded ('x', 'y', or 'z').",
    )

    def _project_axis_x(self, z):
        # x, y, z -> y, z, x
        poly = sg.Polygon(self.poly)
        y_min, _ = self.poly.min(0)
        y_max, _ = self.poly.max(0)
        line = sg.LineString([(y_min, z), (y_max, z)])
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning, module="shapely")
            intersections = poly.intersection(line)

        if not isinstance(intersections, sg.MultiLineString):
            intersection_array = np.asarray(intersections.coords)
            if not intersection_array.shape[0]:
                return []
            intersections = sg.MultiLineString([intersections])

        geoms_2d = []
        for intersection in intersections.geoms:
            intersection = np.asarray(intersection.coords)
            if not intersection.shape[0]:
                continue
            (y_min, _), (y_max, _) = intersection
            y_min, y_max = min(y_min, y_max), max(y_min, y_max)
            x_min, x_max = min(self.h_min, self.h_max), max(self.h_min, self.h_max)
            rect = Rectangle(
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
            )
            geoms_2d.append(rect)
        return geoms_2d

    def _project_axis_y(self, z):
        # x, y, z -> z, x, y
        poly = sg.Polygon(self.poly)
        _, x_min = self.poly.min(0)
        _, x_max = self.poly.max(0)
        line = sg.LineString([(z, x_min), (z, x_max)])
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=RuntimeWarning, module="shapely")
            intersections = poly.intersection(line)

        if not isinstance(intersections, sg.MultiLineString):
            intersection_array = np.asarray(intersections.coords)
            if not intersection_array.shape[0]:
                return []
            intersections = sg.MultiLineString([intersections])

        geoms_2d = []
        for intersection in intersections.geoms:
            intersection = np.asarray(intersection.coords)
            if not intersection.shape[0]:
                continue
            (_, x_min), (_, x_max) = intersection
            x_min, x_max = min(x_min, x_max), max(x_min, x_max)
            y_min, y_max = min(self.h_min, self.h_max), max(self.h_min, self.h_max)
            rect = Rectangle(
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
            )
            geoms_2d.append(rect)
        return geoms_2d

    def _project_axis_z(self, z):
        # x, y, z -> x, y, z
        if z < self.h_min or z < self.h_max:
            return []
        else:
            return [self.poly]

    def _project(self, z):
        if self.axis == "x":
            return self._project_axis_x(z)
        elif self.axis == "y":
            return self._project_axis_y(z)
        else:
            return self._project_axis_z(z)

    def _lumadd(self, sim, material_name, mesh_order, unit=1e-6, xyz="yzx"):
        name = token_hex(4)

        if xyz not in ("xyz", "yzx", "zxy"):
            raise ValueError(
                f"Prism axes should be positively oriented when adding to Lumerical. Got: {xyz!r}"
            )

        sim.addpoly(
            name=name,
            material=material_name,
            override_mesh_order_from_material_database=True,
            mesh_order=mesh_order,
            use_relative_coordinates=True,
            vertices=np.asarray(self.poly, float) * float(unit),
            x=0.0,
            y=0.0,
            z_min=float(self.h_min * unit),
            z_max=float(self.h_max * unit),
        )

        x, y, z = xyz
        if self.axis == "x":
            z, x, y = x, y, z
            # raise NotImplementedError(
            #    "Only prisms extruded perpendicular to the 'chip surface' are currently supported in Lumerical."
            # )
        elif self.axis == "y":
            y, z, x = x, y, z
        xyz = f"{x}{y}{z}"

        if xyz in ["yzx", "zxy"]:
            raise NotImplementedError(
                "Only prisms extruded perpendicular to the 'chip surface' are currently supported in Lumerical."
            )

    def _trimesh(self, color=None, scale=None):
        from trimesh.creation import extrude_polygon  # fmt: skip

        poly = sg.Polygon(self.poly)
        prism = extrude_polygon(poly, self.h_max - self.h_min)
        prism = prism.apply_translation((0, 0, self.h_min))
        if self.axis == "x":
            prism.vertices = np.roll(prism.vertices, shift=1, axis=1)  # type: ignore
        if self.axis == "y":
            prism.vertices = np.roll(prism.vertices, shift=-1, axis=1)  # type: ignore
        if scale is not None:
            sx, sy, sz = scale
            prism.vertices *= np.array([[sx, sy, sz]])  # type: ignore
        if color is not None:
            prism.visual.face_colors = _to_rgba(color)  # type: ignore
        return prism

    def _center(self):
        import shapely.geometry as sg  # fmt: skip

        a, b = np.array(sg.Polygon(self.poly).centroid.xy).ravel()
        c = 0.5 * (self.h_min + self.h_max)
        x, y, z = {
            "x": (c, a, b),
            "y": (b, c, a),
            "z": (a, b, c),
        }[self.axis]
        return x, y, z

    def _axis_tuple(self):
        return {
            "x": (1, 0, 0),
            "y": (0, 1, 0),
            "z": (0, 0, 1),
        }[self.axis]


def _to_rgba(c):
    from matplotlib.colors import to_rgba as _to_rgba_mpl  # fmt: skip
    r, g, b, a = _to_rgba_mpl(c)
    a = min(max(a, 0.1), 0.9)
    return float(r), float(g), float(b), float(a)



""" A CrossSection """

from typing import List, Optional

import numpy as np
from pydantic.v1 import Field

from .base_model import BaseModel, _array, cached_property
from .cell import Cell, _create_full_material_array, _sort_structures
from .environment import Environment
from .mesh import Mesh2D
from .structures import Structure2D


class CrossSection(BaseModel):
    """A `CrossSection` is created from the association of a `Cell` with an `Environment`,
    which uniquely defines the refractive index everywhere."""

    structures: List[Structure2D] = Field(
        description="the 2D structures in the CrossSection"
    )
    mesh: Mesh2D = Field(description="the mesh to discretize the structures with")
    env: Environment = Field(
        description="the environment for which the cross sectionw was calculated"
    )

    @classmethod
    def from_cell(cls, *, cell: Cell, env: Environment):
        cs = cls(structures=cell.structures_2d, mesh=cell.mesh, env=env)
        cs._cache["cell"] = cell
        return cs

    @cached_property
    def materials(self):
        materials = {}
        for i, structure in enumerate(_sort_structures(self.structures), start=1):
            if not structure.material in materials:
                materials[structure.material] = i
        return materials

    @cached_property
    def n_full(self):
        m_full = _create_full_material_array(self.mesh, self.structures, self.materials)
        n_full = np.ones_like(self.mesh.X_full)
        for material, idx in self.materials.items():
            n_full = np.where(m_full == idx, material(self.env), n_full)
        return n_full.view(_array)

    @property
    def nx(self):
        return self.n_full[1::2, ::2].view(_array)

    @property
    def ny(self):
        return self.n_full[::2, 1::2].view(_array)

    @property
    def nz(self):
        return self.n_full[::2, ::2].view(_array)

    def _visualize(self, ax=None, n_cmap=None, cbar=True, show=True, **kwargs):
        import matplotlib.pyplot as plt  # fmt: skip
        from matplotlib import colors  # fmt: skip
        from mpl_toolkits.axes_grid1 import make_axes_locatable  # fmt: skip
        debug_grid = kwargs.pop("debug_grid", False)
        if n_cmap is None:
            n_cmap = colors.LinearSegmentedColormap.from_list(
                name="c_cmap", colors=["#ffffff", "#86b5dc"]
            )
        if ax is not None:
            plt.sca(ax)
        else:
            ax = plt.gca()
        n_full = np.real(self.n_full).copy()
        n_full[0, 0] = 1.0
        plt.pcolormesh(self.mesh.X_full, self.mesh.Y_full, n_full, cmap=n_cmap)
        plt.axis("scaled")
        if not debug_grid:
            plt.grid(True)
        else:
            dx = self.mesh.dx
            dy = self.mesh.dy
            x_ticks = np.sort(np.unique(self.mesh.X_full.ravel()))[::2]
            y_ticks = np.sort(np.unique(self.mesh.Y_full.ravel()))[::2]
            plt.xticks(x_ticks - 0.25 * dx, [f"" for x in x_ticks - 0.25 * dx])
            plt.yticks(y_ticks - 0.25 * dy, [f"" for y in y_ticks - 0.25 * dy])
            plt.xticks(
                x_ticks + 0.25 * dx, [f"" for x in x_ticks + 0.25 * dx], minor=True
            )
            plt.yticks(
                y_ticks + 0.25 * dy, [f"" for y in y_ticks + 0.25 * dy], minor=True
            )
            plt.grid(True, which="major", ls="-")
            plt.grid(True, which="minor", ls=":")
        if cbar:
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            values = np.unique(self.n_full)
            _cbar = plt.colorbar(ticks=values, cax=cax)
            # material_names = ['air'] + [mat.name for mat in self.cell.materials]
            # labels = [f"\n{n}\n{v:.3f}" for n, v in zip(material_names, values)]
            labels = [f"{v:.3f}" for v in values]
            _cbar.ax.set_yticklabels(labels, rotation=90, va="center", ha="center")
            plt.sca(ax)
        if show:
            plt.show()

    @property
    def _cell(self) -> Optional[Cell]:
        """this is a hack. Don't use this property unless you know what you're doing."""
        if "cell" in self._cache:
            return self._cache["cell"]
        else:
            return None



""" a Structure is a combination of a Geometry with a material (and an optional mesh order) """

from typing import Dict, List, Tuple, Union, overload

import numpy as np
from pydantic.v1 import Field

from .base_model import BaseModel
from .geometries import Geometry2D, Geometry3D
from .materials import Material

DEFAULT_MESH_ORDER = 5


def Structure(
    *,
    material: Material,
    geometry: Union[Geometry2D, Geometry3D],
    mesh_order: int = DEFAULT_MESH_ORDER,
):
    kwargs = {
        "material": material,
        "geometry": geometry,
        "mesh_order": mesh_order,
    }
    if isinstance(geometry, Geometry2D):
        return Structure2D(**kwargs)
    else:
        return Structure3D(**kwargs)


class Structure2D(BaseModel):
    """a `Structure2D` is an association between a `Geometry2D` and a `Material`"""

    material: Material = Field(description="the material of the structure")
    geometry: Geometry2D = Field(description="the geometry of the structure")
    mesh_order: int = Field(
        default=DEFAULT_MESH_ORDER, description="the mesh order of the structure"
    )

    def _visualize(self):
        color = self.material.meta.get("color", None)
        return self.geometry._visualize(color=color)


class Structure3D(BaseModel):
    """a `Structure3D` is an association between a `Geometry3D` and a `Material`"""

    material: Material = Field(description="the material of the structure")
    geometry: Geometry3D = Field(description="the geometry of the structure")
    mesh_order: int = Field(
        default=DEFAULT_MESH_ORDER, description="the mesh order of the structure"
    )

    def _project(self, z) -> List[Structure2D]:
        geometry_2d = self.geometry._project(z)
        structs = []
        for geom in geometry_2d:
            struct = Structure2D(
                material=self.material,
                geometry=geom,
                mesh_order=self.mesh_order,
            )
            structs.append(struct)
        return structs

    def _lumadd(self, sim, env, unit=1e-6, xyz="yzx"):
        material_name = self.material._lumadd(sim, env, unit)
        self.geometry._lumadd(sim, material_name, self.mesh_order, unit, xyz)

    def _trimesh(self, color=None, scale=None):
        return self.geometry._trimesh(
            color=(color or self.material.meta.get("color")),
            scale=scale,
        )

    def _visualize(self, scale=None):
        return self._trimesh(scale=scale).show()


def _visualize_structures(structures: List[Structure3D], scale=None):
    """easily visualize a collection (list) of `Structure3D` objects"""
    from trimesh.scene import Scene  # fmt: skip
    from trimesh.transformations import rotation_matrix  # fmt: skip

    scene = Scene(
        geometry=[s._trimesh(scale=scale) for s in _sort_structures(structures)]
    )
    scene.apply_transform(rotation_matrix(np.pi - np.pi / 6, (0, 1, 0)))
    return scene.show()


@overload
def _sort_structures(structures: List[Structure3D]) -> List[Structure3D]:
    ...


@overload
def _sort_structures(structures: List[Structure2D]) -> List[Structure2D]:
    ...


def _sort_structures(
    structures: Union[List[Structure3D], List[Structure2D]]
) -> Union[List[Structure2D], List[Structure3D]]:
    struct_info = [(s.mesh_order, -i, s) for i, s in enumerate(structures)]
    sorted_struct_info = sorted(struct_info, key=lambda I: (I[0], I[1]), reverse=True)
    return [s for _, _, s in sorted_struct_info]  # type: ignore


@overload
def _classify_structures_by_mesh_order_and_material(
    structures: List[Structure3D], materials: Dict[Material, int]
) -> Dict[Tuple[int, int], List[Structure3D]]:
    ...


@overload
def _classify_structures_by_mesh_order_and_material(
    structures: List[Structure2D], materials: Dict[Material, int]
) -> Dict[Tuple[int, int], List[Structure2D]]:
    ...


def _classify_structures_by_mesh_order_and_material(
    structures: Union[List[Structure3D], List[Structure2D]],
    materials: Dict[Material, int],
) -> Union[
    Dict[Tuple[int, int], List[Structure2D]], Dict[Tuple[int, int], List[Structure3D]]
]:
    structures = _sort_structures(structures)
    structures_dict = {}
    for structure in structures:
        mo = structure.mesh_order
        mat = materials[structure.material]
        if (mo, mat) not in structures_dict:
            structures_dict[mo, mat] = []
        structures_dict[mo, mat].append(structure)
    return structures_dict
